from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestBallSave(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_save/'

    def get_platform(self):
        return 'smart_virtual'

    def test_ball_save_enable_in_mode(self):
        # start mode
        self.post_event("start_mode1")

        # mode loaded. mode_ball_save should be enabled
        self.assertTrue(self.machine.ball_saves["mode_ball_save"].enabled)

        # stop mode
        self.post_event("stop_mode1")

        # mode stopped. mode_ball_save should be disabled
        self.assertFalse(self.machine.ball_saves["mode_ball_save"].enabled)

    def test_early_ball_save_once(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        # start game
        self.start_game()
        self.post_event("enable1")

        # ball save should be enabled now
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # early ball save
        self.hit_and_release_switch("s_left_outlane")

        # should eject a second ball
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertBallsOnPlayfield(2)
        self.assertFalse(self.machine.ball_saves["default"].enabled)
        self.assertBallsInPlay(1)

        # second ball drains
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # game should not end and ball should not come back
        self.assertNotEqual(None, self.machine.game)

        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)

        # second ball draining should end the game
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertEqual(None, self.machine.game)

    def test_early_ball_save_unlimited(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        # start game
        self.start_game()
        self.post_event("enable2")

        # ball save should be enabled now
        self.assertTrue(self.machine.ball_saves["unlimited"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)
        self.assertTrue(self.machine.ball_saves["unlimited"].enabled)

        # early ball save
        self.hit_and_release_switch("s_left_outlane")

        # should eject a second ball
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertBallsOnPlayfield(2)
        self.assertTrue(self.machine.ball_saves["unlimited"].enabled)
        self.assertBallsInPlay(1)

        # second ball drains
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # game should not end and ball should not come back
        self.assertNotEqual(None, self.machine.game)

        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)

        # second ball draining should be saved
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertNotEqual(None, self.machine.game)
        self.assertBallsOnPlayfield(0)

        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(1)

    def testBallSaveShootAgain(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.assertFalse(self.machine.ball_saves["default"].enabled)

        # start game
        self.start_game()
        self.post_event("enable1")

        # ball save should be enabled now
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # ball drains right away
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(1)

        # game should not end
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(0, self.machine.playfield.balls)

        # ball save disables because no ball to be saved remain
        self.assertFalse(self.machine.ball_saves["default"].enabled)

        # game should eject another ball
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # ball drains also
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def hurry_up(self, **kwargs):
        self._hurry_up = True

    def grace_period(self, **kwargs):
        self._grace_period = True

    def testBallSaveEvents(self):
        self.machine.events.add_handler("ball_save_default_hurry_up", self.hurry_up)
        self.machine.events.add_handler("ball_save_default_grace_period", self.grace_period)
        self.mock_event("ball_save_default_timer_start")
        self._hurry_up = False
        self._grace_period = False

        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.assertFalse(self.machine.ball_saves["default"].enabled)
        self.assertEqual(0, self._events["ball_save_default_timer_start"])

        # start game
        self.start_game()
        self.post_event("enable1")

        # ball save should be enabled now
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(3)
        self.assertEqual(1, self._events["ball_save_default_timer_start"])
        self.assertNotEqual(None, self.machine.game)

        self.assertEqual(1, self.machine.playfield.balls)
        self.assertFalse(self._hurry_up)
        self.assertFalse(self._grace_period)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        self.advance_time_and_run(8)
        self.assertTrue(self._hurry_up)
        self.assertFalse(self._grace_period)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # should be active for another 2s + 2s grace period
        self.assertTrue(self.machine.ball_saves["default"].enabled)
        self.advance_time_and_run(2)
        self.assertTrue(self._grace_period)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # after another 2s it should turn off
        self.advance_time_and_run(2)
        self.assertFalse(self.machine.ball_saves["default"].enabled)

        # game should still run
        self.assertNotEqual(None, self.machine.game)

        # ball drains
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testBallSaveUnlimited(self):
        self.mock_event("ball_save_unlimited_timer_start")

        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.assertFalse(self.machine.ball_saves["unlimited"].enabled)
        self.assertEqual(0, self._events["ball_save_unlimited_timer_start"])

        # start game
        self.start_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(10)
        self.assertEqual(0, self._events["ball_save_unlimited_timer_start"])

        # ball save should be enabled now
        self.post_event("enable2")
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.ball_saves["unlimited"].enabled)
        self.assertTrue(self.machine.ball_saves["unlimited"].unlimited_saves)
        self.assertEqual(1, self._events["ball_save_unlimited_timer_start"])

        # 20s loop
        for i in range(4):
            # ball drains
            self.machine.switch_controller.process_switch('s_ball_switch1', 1)
            self.machine.switch_controller.process_switch('s_ball_switch2', 1)
            self.advance_time_and_run(1)
            self.assertEqual(0, self.machine.playfield.balls)

            # wait for new ball
            self.advance_time_and_run(4)
            self.assertNotEqual(None, self.machine.game)
            self.assertEqual(1, self.machine.playfield.balls)

        # after 30s + 2s + 2s it should turn off
        self.advance_time_and_run(15)
        self.assertFalse(self.machine.ball_saves["unlimited"].enabled)

        # game should still run
        self.assertNotEqual(None, self.machine.game)

        # ball drains
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testBallOutsideGame(self):
        self.assertEqual(None, self.machine.game)

        # enable ball save
        self.post_event("enable1")
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # ball drains
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.playfield.balls)

        # it should not come back
        self.advance_time_and_run(20)
        self.assertEqual(None, self.machine.game)
        self.assertEqual(0, self.machine.playfield.balls)

    def testBallDoubleDrain(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.assertFalse(self.machine.ball_saves["default"].enabled)

        # start game
        self.start_game()
        self.post_event("enable1")

        self.machine.playfield.add_ball()
        self.advance_time_and_run(4)

        # ball save should be enabled now
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertTrue(self.machine.ball_saves["default"].enabled)

        # double drain
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(1)

        # game should end
        self.assertEqual(None, self.machine.game)
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # first one ball is saved (but will drain soon)
        self.assertEqual(1, self.machine.ball_devices["bd_trough"].balls)
        self.advance_time_and_run(10)

    def test_eject_delay(self):
        # prepare game
        self.fill_troughs()
        self.start_game()
        self.post_event("enable4")
        self.advance_time_and_run()
        self.assertBallNumber(1)
        # one ball on pf
        self.assertEqual(1, self.machine.playfield.available_balls)

        # ball drains
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        # still the same ball
        self.assertBallNumber(1)

        # but no ball on pf
        self.assertEqual(0, self.machine.playfield.available_balls)

        # eject after 10s
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.playfield.available_balls)

    def test_only_last(self):
        # prepare game
        self.fill_troughs()
        self.start_game()
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run(10)
        self.machine.game.balls_in_play = 2

        self.post_event("enable3")
        self.advance_time_and_run()
        self.assertBallNumber(1)
        # two balls on pf
        self.assertEqual(2, self.machine.playfield.available_balls)

        # ball drains
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.advance_time_and_run(1)

        # still the same ball. one ball on pf. no save
        self.assertBallNumber(1)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # last ball drains
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(1)

        # should be safed
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.playfield.available_balls)
        self.assertBallNumber(1)

    def test_unlimited_delay(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run()

        # start game
        self.start_game()
        self.advance_time_and_run(10)
        self.assertBallNumber(1)

        # enable ball save
        self.post_event("enable5")

        # drain ball
        self.drain_all_balls()
        self.advance_time_and_run()

        # ball should not end
        self.assertBallNumber(1)

        # no ball yet on playfield
        self.assertAvailableBallsOnPlayfield(0)

        self.post_event("eject5")
        self.advance_time_and_run()

        # ball 1 continues
        self.assertAvailableBallsOnPlayfield(1)
        self.assertBallNumber(1)

    def test_unlimited_delay_mode(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run()

        # start game
        self.start_game()
        self.advance_time_and_run(10)
        self.assertBallNumber(1)

        # start mode and thereby ball_save
        self.post_event("start_mode2")

        # drain ball
        self.drain_all_balls()
        self.advance_time_and_run()

        # ball should not end
        self.assertBallNumber(1)

        # no ball yet on playfield
        self.assertAvailableBallsOnPlayfield(0)

        self.post_event("mode_ball_save_delayed_eject")
        self.advance_time_and_run()

        # ball 1 continues
        self.assertAvailableBallsOnPlayfield(1)
        self.assertBallNumber(1)

        # save again/drain ball
        self.drain_all_balls()
        self.advance_time_and_run()

        # ball should not end
        self.assertBallNumber(1)

        # no ball yet on playfield
        self.assertAvailableBallsOnPlayfield(0)

        # stop mode
        self.post_event("stop_mode2")
        self.advance_time_and_run()

        # ball 1 continues
        self.assertAvailableBallsOnPlayfield(1)
        self.assertBallNumber(1)

        # ball should end now
        self.drain_all_balls()
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

    def test_dynamic_active_time(self):
        # prepare game
        self.fill_troughs()
        self.advance_time_and_run()

        # start game
        self.start_game()
        self.advance_time_and_run(10)
        self.assertBallNumber(1)

        # Set a player value that's very short, like 2s
        self.machine.game.player["save_time"] = 2
        self.post_event('enable6')
        self.advance_time_and_run()
        # Make sure the save is active
        self.assertTrue(self.machine.ball_saves["dynamic_active_time"].enabled)
        # Run out the short time
        self.advance_time_and_run(4)
        # The ball save should be finished now
        self.assertFalse(self.machine.ball_saves["dynamic_active_time"].enabled)

        # Set a player value that's longer and run the same tests
        self.machine.game.player["save_time"] = 10
        self.post_event('enable6')
        self.advance_time_and_run()
        # Make sure the save is active
        self.assertTrue(self.machine.ball_saves["dynamic_active_time"].enabled)
        # Run out the short time
        self.advance_time_and_run(4)
        # The ball save should still be active
        self.assertTrue(self.machine.ball_saves["dynamic_active_time"].enabled)



from mpf.tests.MpfTestCase import MpfTestCase


class TestPlatform(MpfTestCase):

    def get_config_file(self):
        return 'test_platform.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/platform/'

    def get_platform(self):
        return False

    def test_platform_from_device(self):
        # tests that a platform can be added by a device, even if it's not
        # specified in the hardware section

        self.assertEqual(self.machine.switches["switch1"].platform,
                         self.machine.hardware_platforms['smart_virtual'])

        self.assertEqual(self.machine.switches["switch2"].platform,
                         self.machine.hardware_platforms['virtual'])



from mpf.platforms.virtual import VirtualDriver, VirtualLight

from mpf.tests.MpfTestCase import MpfTestCase


class TestDigitalOutputs(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/digital_output/'

    def test_enable_disable(self):
        light = self.machine.digital_outputs["light_output"].hw_driver
        driver = self.machine.digital_outputs["driver_output"].hw_driver
        self.assertIsInstance(driver, VirtualDriver)
        self.assertIsInstance(light, VirtualLight)
        self.assertEqual("1", driver.number)
        self.assertEqual("test_subtype-1", light.number)
        self.assertEqual("disabled", driver.state)
        self.machine.digital_outputs["driver_output"].enable()
        self.assertEqual("enabled", driver.state)
        self.machine.digital_outputs["driver_output"].disable()
        self.assertEqual("disabled", driver.state)

        self.assertEqual(0.0, light.current_brightness)
        self.machine.digital_outputs["light_output"].enable()
        self.assertEqual(1.0, light.current_brightness)
        self.machine.digital_outputs["light_output"].disable()
        self.assertEqual(0.0, light.current_brightness)



# TODO: test remaining actions
# TODO: test empty control_events

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestTimer(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_timer.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/timer/'

    def _timer_start(self, **kwargs):
        del kwargs
        self.started = True

    def _timer_tick(self, **kwargs):
        del kwargs
        self.tick += 1

    def _timer_complete(self, **kwargs):
        del kwargs
        self.started = False

    def test_start_with_game(self):
        self.start_game()
        self.advance_time_and_run()
        self.assertIn(self.machine.modes["mode_with_timers2"],
                      self.machine.mode_controller.active_modes)

        self.assertIn(self.machine.modes["game"],
                      self.machine.mode_controller.active_modes)

    def test_timer_down_outside_of_game(self):
        self.machine.events.add_handler("timer_timer_down_tick", self._timer_tick)
        self.machine.events.add_handler("timer_timer_down_started", self._timer_start)
        self.machine.events.add_handler("timer_timer_down_complete", self._timer_complete)
        self.machine.events.add_handler("timer_timer_down_stopped", self._timer_complete)

        self.assertFalse(self.machine.modes["mode_with_timers"].active)

        self.tick = 0
        self.started = False

        # timer should not start when mode is not running
        self.machine.events.post('start_timer_down')
        self.advance_time_and_run(10)
        self.assertFalse(self.started)
        self.assertEqual(0, self.tick)

        # mode should not start automatically
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertTrue(self.machine.modes["mode_with_timers"].active)
        self.assertFalse(self.started)
        self.assertEqual(0, self.tick)

        self.machine.events.post('stop_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertFalse(self.machine.modes["mode_with_timers"].active)

        # start mode
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertTrue(self.machine.modes["mode_with_timers"].active)
        self.assertFalse(self.started)
        self.assertEqual(0, self.tick)

        timer = self.machine.timers['timer_down']
        self.assertFalse(timer.running)

        # timer should start now
        self.machine.events.post('start_timer_down')
        self.advance_time_and_run(1)
        self.assertTrue(timer.running)

        self.assertTrue(self.started)
        self.assertEqual(1, self.tick)
        self.assertEqual(5, timer.ticks)
        self.advance_time_and_run(.6)
        self.assertTrue(self.started)
        self.assertEqual(2, self.tick)
        self.assertEqual(4, timer.ticks)
        self.post_event("add_timer_down")
        self.assertEqual(6, timer.ticks)
        self.post_event("subtract_timer_down")
        self.assertEqual(4, timer.ticks)
        self.advance_time_and_run(1.5)
        self.assertEqual(3, self.tick)
        self.assertEqual(3, timer.ticks)
        self.post_event("pause_timer_down")
        self.advance_time_and_run(0.5)
        self.assertEqual(3, self.tick)
        self.assertEqual(3, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(3, self.tick)
        self.assertEqual(3, timer.ticks)
        self.advance_time_and_run(1)
        # Resuming the paused timer re-ticks at the last number,
        # so self.ticks increases but timer.ticks does not
        self.assertEqual(4, self.tick)
        self.assertEqual(3, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(5, self.tick)
        self.assertEqual(2, timer.ticks)
        self.advance_time_and_run(1.5)
        self.assertEqual(6, self.tick)
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1.5)
        self.assertEqual(6, self.tick)
        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1.5)
        self.assertEqual(6, self.tick)
        # and complete at some point
        self.assertFalse(self.started)

        # stay off
        self.advance_time_and_run(20)
        self.assertEqual(6, self.tick)
        self.assertFalse(self.started)

        # cannot be start without reset
        self.post_event("start_timer_down")
        self.advance_time_and_run()
        self.assertEqual(0, timer.ticks)
        self.assertFalse(timer.running)

        self.post_event("reset_timer_down")
        self.advance_time_and_run()
        self.assertEqual(5, timer.ticks)
        self.assertFalse(timer.running)

        self.post_event("start_timer_down")
        self.advance_time_and_run()
        self.assertTrue(timer.running)
        self.assertEqual(5, timer.ticks)
        self.advance_time_and_run()
        self.assertEqual(4, timer.ticks)

    def test_change_tick(self):
        self.start_mode("mode_with_timers")
        self.mock_event("timer_timer_change_tick_tick")
        self.advance_time_and_run()
        self.post_event("timer_change_tick_start")
        self.advance_time_and_run(.1)
        self.assertEventCalled("timer_timer_change_tick_tick", 1)
        self.advance_time_and_run(1)
        self.assertEventCalled("timer_timer_change_tick_tick", 2)
        self.post_event("timer_change_tick_event")
        self.assertEventCalled("timer_timer_change_tick_tick", 2)
        self.advance_time_and_run(1)
        self.assertEventCalled("timer_timer_change_tick_tick", 12)

    def test_start_running(self):
        # add a fake player
        self.start_game()
        self.mock_event("timer_timer_start_running_complete")

        # start mode
        self.machine.events.post('start_mode_with_timers')
        self.machine_run()
        timer = self.machine.timers['timer_start_running']
        self.assertTrue(timer.running)
        self.assertEqual(0, timer.ticks)
        self.assertEqual(0, self._events['timer_timer_start_running_complete'])
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(0, self._events['timer_timer_start_running_complete'])
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.assertTrue(timer.running)
        self.advance_time_and_run(1)
        self.assertFalse(timer.running)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._events['timer_timer_start_running_complete'])

    def test_restart_on_complete(self):
        # add a fake player
        self.start_game()

        # start mode
        self.machine.events.post('start_mode_with_timers')
        self.machine_run()
        timer = self.machine.timers['timer_restart_on_complete']
        self.assertTrue(timer.running)
        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(2, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(3, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(4, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(2, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(3, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(4, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(0, timer.ticks)

        self.assertTrue(timer.running)

    def test_timer_events(self):
        # add a fake player
        self.start_game()

        # start mode
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertTrue(self.machine.modes["mode_with_timers"].active)

        timer = self.machine.timers['timer_up']
        self.assertFalse(timer.running)

        # timer should start now
        self.post_event('start_timer_up')
        self.assertTrue(timer.running)

        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(2, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(3, timer.ticks)
        self.post_event('reset_timer_up')

        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.post_event('stop_timer_up')
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.post_event('restart_timer_up')
        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)
        self.post_event('jump_timer_up')
        self.assertEqual(5, timer.ticks)
        self.post_event_with_params("change_tick_interval_timer_up", change=1)
        # 1s * 4 = 4s
        self.advance_time_and_run(1)
        self.assertEqual(5, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(5, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(5, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(6, timer.ticks)
        self.advance_time_and_run(4)
        self.assertEqual(7, timer.ticks)
        self.post_event("change_tick_interval_timer_up")
        # 4s * 4 = 16s
        self.advance_time_and_run(8)
        self.assertEqual(7, timer.ticks)
        self.advance_time_and_run(8)
        self.assertEqual(8, timer.ticks)

        self.post_event("set_tick_interval_timer_up")
        # back to 2s
        self.advance_time_and_run(1)
        self.assertEqual(8, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(9, timer.ticks)
        self.advance_time_and_run(2)
        self.assertEqual(10, timer.ticks)
        # and complete at some point
        self.assertFalse(timer.running)

        self.post_event('jump_timer_up')
        self.assertEqual(5, timer.ticks)
        self.post_event('jump_over_max_timer_up')
        self.assertEqual(15, timer.ticks)
        self.post_event('add_timer_up')
        self.assertEqual(15, timer.ticks)

        self.post_event('restart_timer_up')
        self.post_event("reset_tick_interval")
        self.advance_time_and_run()
        self.assertEqual(1, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(2, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(3, timer.ticks)
        self.post_event("change_tick_interval_timer_up")
        self.advance_time_and_run(4)
        self.assertEqual(4, timer.ticks)
        self.post_event("reset_tick_interval")
        self.advance_time_and_run()
        self.assertEqual(5, timer.ticks)

    def test_interrupt_timer_by_mode_stop_with_player(self):
        self.machine.events.add_handler("timer_timer_down_tick", self._timer_tick)
        self.machine.events.add_handler("timer_timer_down_started", self._timer_start)
        self.machine.events.add_handler("timer_timer_down_complete", self._timer_complete)
        self.machine.events.add_handler("timer_timer_down_stopped", self._timer_complete)

        # add a fake player
        self.start_game()

        self.assertFalse(self.machine.modes["mode_with_timers"].active)

        self.tick = 0
        self.started = False

        # start mode
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertTrue(self.machine.modes["mode_with_timers"].active)
        self.assertFalse(self.started)
        self.assertEqual(0, self.tick)

        # timer should start now
        self.machine.events.post('start_timer_down')
        self.advance_time_and_run(1)

        self.assertTrue(self.started)
        self.assertEqual(1, self.tick)
        self.advance_time_and_run(.6)
        self.assertTrue(self.started)
        self.assertEqual(2, self.tick)
        self.advance_time_and_run(1.5)
        self.assertEqual(3, self.tick)

        # stop mode. timer should stop
        self.machine.events.post('stop_mode_with_timers')
        self.advance_time_and_run(10)
        self.assertFalse(self.machine.modes["mode_with_timers"].active)
        self.assertFalse(self.started)

        # and stay off
        self.advance_time_and_run(20)
        self.assertEqual(3, self.tick)
        self.assertFalse(self.started)

    def test_mode_timer_with_player_var(self):
        # add a fake player
        self.start_game()

        # start mode. no player vars
        self.mock_event("timer_timer_player_var_complete")
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(.1)

        self.assertEventCalled("timer_timer_player_var_complete")
        self.machine.events.post('stop_mode_with_timers')
        self.advance_time_and_run(.1)

        # set player vars. timer should run 5s
        self.machine.game.player.start = 2
        self.machine.game.player.end = 7

        self.machine.log.debug("START")
        self.mock_event("timer_timer_player_var_complete")
        self.machine.events.post('start_mode_with_timers')
        self.advance_time_and_run(4.5)
        self.machine.log.debug("END")
        self.assertEventNotCalled("timer_timer_player_var_complete")

        self.advance_time_and_run(0.6)

        self.assertEventCalled("timer_timer_player_var_complete")
        self.machine.events.post('stop_mode_with_timers')

    def test_timer_control_with_player_var(self):
        self.start_game()
        self.machine.game.player.timer_amount = 10
        timer = self.machine.timers['timer_with_player_var_control_events']

        self.machine.events.post('start_player_var_timer')
        self.advance_time_and_run(.1)
        self.assertTrue(timer.running)
        self.assertEqual(0, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(1, timer.ticks)

        self.machine.events.post('add_player_var_timer')
        self.advance_time_and_run(.1)
        self.assertEqual(11, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(12, timer.ticks)

        self.machine.game.player.timer_amount = 5
        self.machine.events.post('subtract_player_var_timer')
        self.advance_time_and_run(.1)
        self.assertEqual(7, timer.ticks)
        self.advance_time_and_run(1)
        self.assertEqual(8, timer.ticks)



from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase


class TestBlockingEvents(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/blocking_events/'

    def testBlockingInMode(self):
        self.handler = MagicMock()
        self.handler2 = MagicMock()
        self.handler3 = MagicMock()
        self.machine.events.add_handler("event1", self.handler, priority=10, blocking_facility="random")
        self.machine.events.add_handler("event1", self.handler2, priority=150, blocking_facility="random")
        self.machine.events.add_handler("event1", self.handler3, priority=9)

        self.post_event("event1")
        self.advance_time_and_run(.1)
        self.handler.assert_called_with()
        self.handler2.assert_called_with()
        self.handler3.assert_called_with()
        self.handler.reset_mock()
        self.handler2.reset_mock()
        self.handler3.reset_mock()

        self.start_mode("mode1")
        self.advance_time_and_run(.1)
        self.post_event("event1")
        self.advance_time_and_run(.1)
        self.handler.assert_not_called()
        self.handler2.assert_called_with()
        self.handler3.assert_called_with(_min_priority={'all': 100})
        self.handler.reset_mock()
        self.handler2.reset_mock()
        self.handler3.reset_mock()

        self.start_mode("mode2")
        self.advance_time_and_run(.1)
        self.post_event("event1")
        self.advance_time_and_run(.1)
        self.handler.assert_not_called()
        self.handler2.assert_not_called()
        self.handler3.assert_called_with(_min_priority={'all': 200})
        self.handler.reset_mock()
        self.handler2.reset_mock()
        self.handler3.reset_mock()

        self.stop_mode("mode2")
        self.advance_time_and_run(.1)
        self.post_event("event1")
        self.advance_time_and_run(.1)
        self.handler.assert_not_called()
        self.handler2.assert_called_with()
        self.handler3.assert_called_with(_min_priority={'all': 100})




"""Implements enable and disable events for devices."""
from typing import Optional

import abc

from mpf.core.system_wide_device import SystemWideDevice

from mpf.core.device_monitor import DeviceMonitor

from mpf.core.events import event_handler

from mpf.core.mode_device import ModeDevice
from mpf.core.machine import MachineController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.mode import Mode  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.player import Player  # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("enabled")
class EnableDisableMixin(ModeDevice, metaclass=abc.ABCMeta):

    """Implements enable and disable_events."""

    __slots__ = ["_enabled", "player"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Remember the enable state."""
        self._enabled = None    # type: Optional[bool]
        self.player = None      # type: Optional[Player]
        super().__init__(machine, name)

    def _enable(self):
        """Enable the device.

        This can be caused by a mode restore, initial enable at boot or by an
        enable_event.
        """

    def _disable(self):
        """Disable the device.

        This can be caused by a mode stop or by an disable_event.
        """

    @event_handler(10)
    def event_enable(self, **kwargs):
        """Handle enable control event."""
        del kwargs
        self.enable()

    def enable(self) -> None:
        """Enable device."""
        if self.enabled is True:
            return
        self.enabled = True
        self.notify_virtual_change("enabled", False, True)      # type: ignore
        self._enable()

    def add_control_events_in_mode(self, mode) -> None:
        """Remove enable here."""

    @event_handler(0)
    def event_disable(self, **kwargs):
        """Handle disable control event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable device."""
        if self.enabled is False:
            return
        self.enabled = False
        self.notify_virtual_change("enabled", True, False)
        self._disable()

    def _load_enable_based_on_config_default(self) -> None:
        """Load default enable state from config."""
        if 'start_enabled' in self.config and self.config['start_enabled'] is True:
            start_enabled = True
        elif 'start_enabled' in self.config and self.config['start_enabled'] is False:
            start_enabled = False
        elif 'enable_events' in self.config:
            start_enabled = not self.config['enable_events']
        else:
            start_enabled = False

        if start_enabled:
            self._enable()
            self.enabled = True
        else:
            self.enabled = False

    @property
    def enabled(self):
        """Return true if enabled."""
        if 'persist_enable' in self.config and self.config['persist_enable']:
            # in case the mode is not running the device is disabled
            if not self.player:
                return False
            return self.player["{}_{}_enabled".format(self.class_label, self.name)]

        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """Set enabled enabled."""
        if 'persist_enable' in self.config and self.config['persist_enable']:
            self.player["{}_{}_enabled".format(self.class_label, self.name)] = value
        else:
            self._enabled = value

    @property
    def persist_enabled(self):
        """Return if enabled is persisted."""
        return 'persist_enable' in self.config and self.config['persist_enable']

    def device_loaded_in_mode(self, mode: "Mode", player) -> None:
        """Check enable on mode start."""
        super().device_loaded_in_mode(mode, player)
        self.player = player
        if self.persist_enabled:
            if not player.is_player_var("{}_{}_enabled".format(self.class_label, self.name)):
                self._load_enable_based_on_config_default()
            elif self.enabled:
                self._enable()
        else:
            self._load_enable_based_on_config_default()

    def device_removed_from_mode(self, mode) -> None:
        """Forget enable state."""
        del mode
        self._disable()
        self.player = None
        self._enabled = None


@DeviceMonitor("enabled")
class EnableDisableMixinSystemWideDevice(SystemWideDevice, metaclass=abc.ABCMeta):

    """Implements enable and disable_events."""

    __slots__ = ["enabled"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Remember the enable state."""
        self.enabled = False    # type: bool
        super().__init__(machine, name)

    def _enable(self):
        """Enable the device.

        Overwrite this method.
        """

    def _disable(self):
        """Disable the device.

        Overwrite this method.
        """

    @event_handler(10)
    def event_enable(self, **kwargs):
        """Handle enable control event."""
        del kwargs
        self.enable()

    def enable(self) -> None:
        """Enable device."""
        if self.enabled is True:
            return
        self.enabled = True
        self._enable()

    @event_handler(0)
    def event_disable(self, **kwargs):
        """Handle disable control event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable device."""
        if self.enabled is False:
            return
        self.enabled = False
        self._disable()



import serial

class TMCLError(Exception):
    pass

def encodeRequestCommand(m_address, n_command, n_type, n_motor, value, debug=False):
    # convert to bytes
    m_address = int(m_address) % (1<<8)
    n_command = int(n_command) % (1<<8)
    n_type = int(n_type) % (1<<8)
    n_motor = int(n_motor) % (1<<8)
    value = [(int(value) >> i*8) % (1<<8) for i in range(3,-1,-1)]
    # generate command
    checksum = (m_address + n_command + n_type + n_motor + sum(value)) % (1<<8)
    tmcl_bytes = [m_address, n_command, n_type, n_motor] + value + [checksum]
    tmcl_cmd = sum(b << (8-i)*8 for i,b in enumerate(tmcl_bytes))
    #if debug:
    #    print "{0:>18X}".format(tmcl_cmd), "".join([chr(b) for b in tmcl_bytes])
    return "".join([chr(b) for b in tmcl_bytes])

def encodeReplyCommand(r_address, m_address, status, n_command, value, debug=False):
    # convert to bytes
    r_address = int(r_address) % (1<<8)
    m_address = int(m_address) % (1<<8)
    status = int(status) % (1<<8)
    n_command = int(n_command) % (1<<8)
    value = [(int(value) >> i*8) % (1<<8) for i in range(3,-1,-1)]
    # generate command
    checksum = (r_address + m_address + status + n_command + sum(value)) % (1<<8)
    tmcl_bytes = [r_address, m_address, status, n_command] + value + [checksum]
    tmcl_cmd = sum(b << (8-i)*8 for i,b in enumerate(tmcl_bytes))
    #if debug:
    #    print "{0:>18X}".format(tmcl_cmd), "".join([chr(b) for b in tmcl_bytes])
    return "".join([chr(b) for b in tmcl_bytes])

def decodeRequestCommand(cmd_string):
    byte_array = bytearray(cmd_string)
    if len(byte_array) != 9:
        raise TMCLError("Commandstring shorter than 9 bytes")
    if byte_array[8] != sum(byte_array[:8]) % (1<<8):
        raise TMCLError("Checksum error in command %s" % cmd_string)
    ret = {'module-address': byte_array[0], 'command-number': byte_array[1], 'type-number': byte_array[2],
           'motor-number': byte_array[3], 'value': sum(b << (3 - i) * 8 for i, b in enumerate(byte_array[4:8])),
           'checksum': byte_array[8]}
    return ret

def decodeReplyCommand(cmd_string):
    byte_array = bytearray(cmd_string)
    if len(byte_array) != 9:
        raise TMCLError("Commandstring shorter than 9 bytes")
    if byte_array[8] != sum(byte_array[:8]) % (1<<8):
        raise TMCLError("Checksum error in command %s" % cmd_string)
    ret = {'reply-address': byte_array[0], 'module-address': byte_array[1], 'status': byte_array[2],
           'command-number': byte_array[3], 'value': sum(b << (3 - i) * 8 for i, b in enumerate(byte_array[4:8])),
           'checksum': byte_array[8]}
    return ret

STATUSCODES = { 100 : "Succesfully executed, no error",
                101 : "Command loaded into TMCL program EEPROM",
                  1 : "Wrong Checksum",
                  2 : "Invalid command",
                  3 : "Wrong type",
                  4 : "Invalid value",
                  5 : "Configuration EEPROM locked",
                  6 : "Command not available" }

STAT_OK = 100

COMMAND_NUMBERS = {  1 : "ROR",    2 : "ROL",    3 : "MST",
                     4 : "MVP",    5 : "SAP",    6 : "GAP",
                     7 : "STAP",   8 : "RSAP",   9 : "SGP",
                    10 : "GGP",   11 : "STGP",  12 : "RSGP",
                    13 : "RFS",   14 : "SIO",   15 : "GIO",
                    19 : "CALC",  20 : "COMP",  21 : "JC",
                    22 : "JA",    23 : "CSUB",  24 : "RSUB",
                    25 : "EI",    26 : "DI",    27 : "WAIT",
                    28 : "STOP",  30 : "SCO",   31 : "GCO",
                    32 : "CCO",   33 : "CALCX", 34 : "AAP",
                    35 : "AGP",   37 : "VECT",  38 : "RETI",
                    39 : "ACO"
                  }

NUMBER_COMMANDS = dict([(v, k) for k, v in COMMAND_NUMBERS.items()])

INTERRUPT_VECTORS = {  0 : "Timer 0",
                       1 : "Timer 1",
                       2 : "Timer 2",
                       3 : "Target position reached",
                      15 : "stallGuard",
                      21 : "Deviation",
                      27 : "Left stop switch",
                      28 : "Right stop switch",
                      39 : "Input change 0",
                      40 : "Input change 1",
                     255 : "Global interrupts" }

CMD_MVP_TYPES = { 'ABS' : 0,
                  'REL' : 1,
                  'COORD' : 2 }
CMD_RFS_TYPES = { 'START' : 0,
                  'STOP' : 1,
                  'STATUS' : 2 }


def apard(name, prange, acc):
    return { "name": name,
             "range" : prange,
             "access" : acc }

TR_24s = [(-2**23+1, 2**23)]
TR_32u = [(0, 2**32)]
TR_32s = [(-2**31, 2**31)]
TR_16u = [(0, 2**16)]
TR_12s = [(-2**11+1, 2**11)]
TR_12u = [(0, 2**12)]
TR_11u = [(0, 2**11)]
TR_10u = [(0, 2**10)]
TR_8u = [(0, 2**8)]
TR_7s = [(-2**6, 2**6)]
TR_5u = [(0, 2**5)]
TR_1u = [(0, 2**1)]
TR_m3 = [(0, 3)]
TR_m4 = [(0, 4)]
TR_m9 = [(0, 9)]
TR_m12 = [(0, 14)]
TR_m14 = [(0, 14)]
TR_m16 = [(0, 16)]

TR_xCHP0 = [(-3, 13)]
TR_xCHP1 = [(0, 1), (2, 16)]
TR_xSE0 = [(1, 4)]
TR_xRFS0 = [(1, 137)]
TR_xRFS1 = [(0, 8388307)]
TR_xPWR0 = [(1, 2**16)]
TR_xRND0 = [(0, 2**31)]

T_R = 4
T_W = 2
T_E = 1
T_RW = T_R+T_W
T_RWE = T_RW + T_E


AXIS_PARAMETER = {   0 : ("target position", TR_24s, T_RW),
                     1 : ("actual position", TR_24s, T_RW),
                     2 : ("target speed", TR_12s, T_RW),
                     3 : ("actual speed", TR_12s, T_RW),
                     4 : ("max positioning speed", TR_11u, T_RWE),
                     5 : ("max acceleration", TR_11u, T_RWE),
                     6 : ("abs max current", TR_8u, T_RWE),
                     7 : ("standby current", TR_8u, T_RWE),
                     8 : ("target pos reached", TR_1u, T_R),
                     9 : ("ref switch status", TR_1u, T_R),
                    10 : ("right limit switch status", TR_1u, T_R),
                    11 : ("left limit switch status", TR_1u, T_R),
                    12 : ("right limit switch disable", TR_1u, T_RWE),
                    13 : ("left limit switch disable", TR_1u, T_RWE),
                   130 : ("minimum speed", TR_11u, T_RWE),
                   135 : ("actual acceleration", TR_11u, T_R),
                   138 : ("ramp mode", TR_m3, T_RWE),
                   140 : ("microstep resolution", TR_m9, T_RWE),
                   141 : ("ref switch tolerance", TR_12u, T_RW),
                   149 : ("soft stop flag", TR_1u, T_RWE),
                   153 : ("ramp divisor", TR_m14, T_RWE),
                   154 : ("pulse divisor", TR_m14, T_RWE),
                   160 : ("step interpolation enable", TR_1u, T_RW),
                   161 : ("double step enable", TR_1u, T_RW),
                   162 : ("chopper blank time", TR_m4, T_RW),
                   163 : ("chopper mode", TR_1u, T_RW),
                   164 : ("chopper hysteresis dec", TR_m4, T_RW),
                   165 : ("chopper hysteresis end", TR_xCHP0, T_RW),
                   166 : ("chopper hysteresis start", TR_m9, T_RW),
                   167 : ("chopper off time", TR_xCHP1, T_RW),
                   168 : ("smartEnergy min current", TR_1u, T_RW),
                   169 : ("smartEnergy current downstep", TR_m4, T_RW),
                   170 : ("smartEnergy hysteresis", TR_m16, T_RW),
                   171 : ("smartEnergy current upstep", TR_xSE0, T_RW),
                   172 : ("smartEnergy hysteresis start", TR_m16, T_RW),
                   173 : ("stallGuard2 filter enable", TR_1u, T_RW),
                   174 : ("stallGuard2 threshold", TR_7s, T_RW),
                   175 : ("slope control high side", TR_m4, T_RW),
                   176 : ("slope control low side", TR_m4, T_RW),
                   177 : ("short protection disable", TR_1u, T_RW),
                   178 : ("short detection timer", TR_m4, T_RW),
                   179 : ("Vsense", TR_1u, T_RW),
                   180 : ("smartEnergy actual current", TR_5u, T_RW),
                   181 : ("stop on stall", TR_11u, T_RW),
                   182 : ("smartEnergy threshold speed", TR_11u, T_RW),
                   183 : ("smartEnergy slow run current", TR_8u, T_RW),
                   193 : ("ref. search mode", TR_xRFS0, T_RWE),
                   194 : ("ref. search speed", TR_11u, T_RWE),
                   195 : ("ref. switch speed", TR_11u, T_RWE),
                   196 : ("distance end switches", TR_xRFS1, T_R),
                   204 : ("freewheeling", TR_16u, T_RWE),
                   206 : ("actual load value", TR_10u, T_R),
                   208 : ("TMC262 errorflags", TR_8u, T_R),
                   209 : ("encoder pos", TR_24s, T_RW),
                   210 : ("encoder prescaler", TR_16u, T_RWE), # that one isnt really correct
                   212 : ("encoder max deviation", TR_16u, T_RWE),
                   214 : ("power down delay", TR_xPWR0, T_RWE)
                }

#SINGLE_AXIS_PARAMETERS = [140]+range(160, 184)



GLOBAL_PARAMETER = { (0, 64) : ("EEPROM magic", TR_8u, T_RWE),
                     (0, 65) : ("RS485 baud rate", TR_m12, T_RWE),
                     (0, 66) : ("serial address", TR_8u, T_RWE),
                     (0, 73) : ("EEPROM lock flag", TR_1u, T_RWE),
                     (0, 75) : ("telegram pause time", TR_8u, T_RWE),
                     (0, 76) : ("serial host adress", TR_8u, T_RWE),
                     (0, 77) : ("auto start mode", TR_1u, T_RWE),
                     (0, 81) : ("TMCL code protect", TR_m4, T_RWE),
      #Wrong type?? #(0, 84) : ("coordinate storage", TR_1u, T_RWE),
                     (0, 128) : ("TMCL application status", TR_m3, T_R),
                     (0, 129) : ("download mode", TR_1u, T_R),
                     (0, 130) : ("TMCL program counter", TR_32u, T_R),
                     (0, 132) : ("tick timer", TR_32u, T_RW),
      #Wrong type?? #(0, 133) : ("random number", TR_xRND0, T_R),
                     (3, 0) : ("Timer0 period", TR_32u, T_RWE),
                     (3, 1) : ("Timer1 period", TR_32u, T_RWE),
                     (3, 2) : ("Timer2 period", TR_32u, T_RWE),
                     (3, 39) : ("Input0 edge type", TR_m4, T_RWE),
                     (3, 40) : ("Input0 edge type", TR_m4, T_RWE)
                   }
# add general purpose registers
for b, p, a in zip([2]*256, range(256), ([T_RWE]*56)+([T_RW]*200)):
    GLOBAL_PARAMETER[(2, p)] = ("general purpose reg#{0:0>3d}".format(p), TR_32s, a)


class TMCLDevice(object):

    def __init__(self, port="/dev/ttyACM0", debug=False):
        self._port = port
        self._debug = debug
        self.serial = serial.Serial(port)

    def stop(self):
        """Close serial."""
        self.serial.close()

    def _query(self, request):
        """Encode and send a query. Receive, decode, and return reply"""
        #Insert inside encode request command function a way to check the value ranges
        req = encodeRequestCommand(*request)
        req = list(map(ord,req))
        if self._debug:
            print(("send to TMCL: ",self._hexString(req),decodeRequestCommand(req)))
        self.serial.write(req)
        resp = decodeReplyCommand(self.serial.read(9))
        if self._debug:
            tmp = list(resp.values())[:-1]
            tmp = encodeReplyCommand(*tmp)
            print(("got from TMCL:",self._hexString(tmp), resp))
        return resp['status'], resp['value']

    def _hexString(self, cmd):
        """Convert encoded command string to human-readable string of hex values"""
        temp = None
        #Quickfix
        if(type(cmd[0]) is str):
            temp = [ord(i) for i in cmd]
        elif(type(cmd[0]) is int):
            temp = [i for i in cmd]

        s = ['{:x}'.format(i).rjust(2) for i in temp]
        return "[" + ", ".join(s) + "]"

    def _pn_checkrange(self, parameter_number, value, prefix):
        pn = parameter_number
        v = int(value)
        DICT = AXIS_PARAMETER if type(pn) == int else GLOBAL_PARAMETER
        if not pn in DICT.keys():
            raise TMCLError(prefix+"parameter number not valid")
        name, ranges, _ = DICT[parameter_number]
        NOTINRANGE = False
        for (l,h) in ranges:
            if not (l <= v < h):
                NOTINRANGE=True
        else:
            if NOTINRANGE: raise TMCLError(prefix+"parameter "+name+" needs "+
                "+".join(["range("+str(l)+", "+str(h)+")" for l, h in ranges]))
        return pn, v

    def ror(self, motor_number, velocity):
        """
        tmcl_ror(motor_number, velocity) --> None

        The motor will be instructed to rotate with a specified velocity
        in right direction (increasing the position counter).

        TMCL-Mnemonic: ROR <motor number>, <velocity>
        """
        cn = NUMBER_COMMANDS['ROR']
        mn = int(motor_number)
        v = int(velocity)
        if not 0 <= mn <= 2:
            raise TMCLError("ROR: motor_number not in range(3)")
        if not 0 <= v <= 2047:
            raise TMCLError("ROR: velocity not in range(2048)")
        status, value = self._query((0x01, cn, 0x00, mn, v))
        if status != STAT_OK:
            raise TMCLError("ROR: got status "+STATUSCODES[status])
        return None

    def rol(self, motor_number, velocity):
        """
        tmcl_rol(motor_number, velocity) --> None

        With this command the motor will be instructed to rotate with a
        specified velocity (opposite direction compared to tmcl_rol,
        decreasing the position counter).

        TMCL-Mnemonic: ROL <motor number>, <velocity>
        """
        cn = NUMBER_COMMANDS['ROL']
        mn = int(motor_number)
        v = int(velocity)
        if not 0 <= mn <= 2:
            raise TMCLError("ROL: motor_number not in range(3)")
        if not 0 <= v <= 2047:
            raise TMCLError("ROL: velocity not in range(2048)")
        status, value = self._query((0x01, cn, 0x00, mn, v))
        if status != STAT_OK:
            raise TMCLError("ROL: got status "+STATUSCODES[status])
        return None

    def mst(self, motor_number):
        """
        tmcl_mst(motor_number) --> None

        The motor will be instructed to stop.

        TMCL-Mnemonic: MST <motor number>
        """
        cn = NUMBER_COMMANDS['MST']
        mn = int(motor_number)
        if not 0 <= mn <= 2:
            raise TMCLError("MST: motor_number not in range(3)")
        status, value = self._query((0x01, cn, 0x00, mn, 0x00))
        if status != STAT_OK:
            raise TMCLError("MST: got status "+STATUSCODES[status])
        return None

    def mvp(self, motor_number, cmdtype, value):
        """
        tmcl_mvp(motor_number, type, value) --> None

        The motor will be instructed to move to a specified relative or
        absolute position or a pre-programmed coordinate. It will use
        the acceleration/deceleration ramp and the positioning speed
        programmed into the unit. This command is non-blocking: that is,
        a reply will be sent immediately after command interpretation
        and initialization of the motion controller. Further commands
        may follow without waiting for the motor reaching its end
        position. The maximum velocity and acceleration are defined by
        axis parameters #4 and #5.

        Three operation types are available:
            * Moving to an absolute position in the range from
              -8388608 to +8388607 (-223 to+223-1).
            * Starting a relative movement by means of an offset to the
              actual position. In this case, the new resulting position
              value must not exceed the above mentioned limits, too.
            * Moving the motor to a (previously stored) coordinate
              (refer to SCO for details).

        TMCL-Mnemonic: MVP <ABS|REL|COORD>, <motor number>,
                           <position|offset|coordinate number>
        """
        cn = NUMBER_COMMANDS['MVP']
        mn = int(motor_number)
        t = str(cmdtype)
        v = int(value)
        if not 0 <= mn <= 2:
            raise TMCLError("MVP: motor_number not in range(3)")
        if t not in CMD_MVP_TYPES.keys():
            raise TMCLError("MVP: type not in ['ABS', 'REL', 'COORD']")
        if t == 'ABS' and not -2**23 <= v <= 2**23:
            raise TMCLError("MVP: ABS: value not in range(-2**23,2**23)")
        # pass 'REL' because we dont know the current pos here
        if t == 'COORD' and not 0 <= v <= 20:
            raise TMCLError("MVP: COORD: value not in range(21)")
        t = CMD_MVP_TYPES[t] % (1<<8)
        status, value = self._query((0x01, cn, t, mn, v))
        if status != STAT_OK:
            raise TMCLError("MVP: got status "+STATUSCODES[status])
        return None

    def rfs(self, motor_number, cmdtype):
        """
        tmcl_rfs(motor_number, cmdtype) --> int

        The TMCM-1110 has a built-in reference search algorithm which
        can be used. The reference search algorithm provides switching
        point calibration and three switch modes. The status of the
        reference search can also be queried to see if it has already
        finished. (In a TMCLTM program it is better to use the WAIT
        command to wait for the end of a reference search.) Please see
        the appropriate parameters in the axis parameter table to
        configure the reference search algorithm to meet your needs
        (chapter 6). The reference search can be started, stopped, and
        the actual status of the reference search can be checked.

        if cmdtype in ['START', 'STOP']:
            return 0
        if cmdtype == 'STATUS':
            return 0 if "ref-search is active" else "other values"

        TMCL-Mnemonic: RFS <START|STOP|STATUS>, <motor number>
        """
        cn = NUMBER_COMMANDS['RFS']
        mn = int(motor_number)
        t = str(cmdtype)
        if not 0 <= mn <= 2:
            raise TMCLError("RFS: motor_number not in range(3)")
        if t not in CMD_RFS_TYPES.keys():
            raise TMCLError("RFS: type not in ['START', 'STOP', 'STATUS']")
        t = CMD_RFS_TYPES[t] % (1<<8)
        status, value = self._query((0x01, cn, t, mn, 0x0000))
        if status != STAT_OK:
            raise TMCLError("RFS: got status "+STATUSCODES[status])
        return value if t == CMD_RFS_TYPES['STATUS'] else 0

    def cco(self, motor_number, coordinate_number):
        """
        tmcl_cco(motor_number, coordinate_number) --> None

        The actual position of the axis is copied to the selected
        coordinate variable. Depending on the global parameter 84, the
        coordinates are only stored in RAM or also stored in the EEPROM
        and copied back on startup (with the default setting the
        coordinates are stored in RAM only). Please see the SCO and GCO
        commands on how to copy coordinates between RAM and EEPROM.
        Note, that the coordinate number 0 is always stored in RAM only.

        TMCL-Mnemonic: CCO <coordinate number>, <motor number>
        """
        cn = NUMBER_COMMANDS['CCO']
        mn = int(motor_number)
        coord_n = int(coordinate_number)
        if not 0 <= mn <= 2:
            raise TMCLError("CCO: motor_number not in range(3)")
        if not 0 <= coord_n <= 20:
            raise TMCLError("CCO: coordinate_number not in range(21)")
        status, value = self._query((0x01, cn, coord_n, mn, 0x0000))
        if status != STAT_OK:
            raise TMCLError("CCO: got status "+STATUSCODES[status])
        return None

    def sco(self, motor_number, coordinate_number, position):
        """
        tmcl_sco(self, motor_number, coordinate_number, position) --> None

        Up to 20 position values (coordinates) can be stored for every
        axis for use with the MVP COORD command. This command sets a
        coordinate to a specified value. Depending on the global
        parameter 84, the coordinates are only stored in RAM or also
        stored in the EEPROM and copied back on startup (with the
        default setting the coordinates are stored in RAM only).
        Please note that the coordinate number 0 is always stored in
        RAM only.

        TMCL-Mnemonic: SCO <coordinate number>, <motor number>, <position>
        """
        cn = NUMBER_COMMANDS['SCO']
        mn = int(motor_number)
        coord_n = int(coordinate_number)
        pos = int(position)
        if not 0 <= coord_n <= 20:
            raise TMCLError("SCO: coordinate_number not in range(21)")
        if not -2**23 <= pos <= 2**23:
            raise TMCLError("SCO: position not in range(-2**23,2**23)")
        if not 0 <= mn <= 2:
            raise TMCLError("SCO: motor_number not in range(3)")
        elif not (mn == 0xFF and pos == 0):
            raise TMCLError("SCO: special function needs pos == 0")
        status, value = self._query((0x01, cn, coord_n, mn, pos))
        if status != STAT_OK:
            raise TMCLError("SCO: got status "+STATUSCODES[status])
        return None

    def gco(self, motor_number, coordinate_number):
        """
        tmcl_gco(self, motor_number, coordinate_number) --> int

        This command makes possible to read out a previously stored
        coordinate. In standalone mode the requested value is copied to
        the accumulator register for further processing purposes such
        as conditioned jumps. In direct mode, the value is only output
        in the value field of the reply, without affecting the
        accumulator. Depending on the global parameter 84, the
        coordinates are only stored in RAM or also stored in the EEPROM
        and copied back on startup (with the default setting the
        coordinates are stored in RAM, only).
        Please note that the coordinate number 0 is always stored in
        RAM, only.

        TMCL-Mnemonic: GCO <coordinate number>, <motor number>
        """
        cn = NUMBER_COMMANDS['GCO']
        mn = int(motor_number)
        coord_n = int(coordinate_number)
        if not 0 <= coord_n <= 20:
            raise TMCLError("GCO: coordinate_number not in range(21)")
        if not (0 <= mn <= 2 or mn == 0xFF):
            raise TMCLError("GCO: motor_number not in range(3)")
        elif not (mn == 0xFF and pos == 0):
            raise TMCLError("GCO: special function needs pos == 0")
        status, value = self._query((0x01, cn, coord_n, mn, pos))
        if status != STAT_OK:
            raise TMCLError("GCO: got status "+STATUSCODES[status])
        return value

    def sio(self, port_number, state):
        """
        tmcl_sio(output_number, state) --> None

        This command sets the status of the general digital output
        either to low (0) or to high (1).

        TMCL-Mnemonic: SIO <port number>, <bank number>, <value>
        """
        cn = NUMBER_COMMANDS['SIO']
        outp = int(port_number)
        s = bool(state)
        if not 0 <= outp <= 4:
            raise TMCLError("SIO: output_number not in range(5)")
        status, value = self._query((0x01, cn, outp, 0x02, s))
        if status != STAT_OK:
            raise TMCLError("SIO: got status "+STATUSCODES[status])
        return None

    def gio(self, port_number, bank_number):
        """
        tmcl_gio(port_number, bank_number) --> int

        With this command the status of the two available general
        purpose inputs of the module can be read out. The function
        reads a digital or analogue input port. Digital lines will read
        0 and 1, while the ADC channels deliver their 12 bit result in
        the range of 0... 4095. In direct mode the value is only output
        in the value field of the reply, without affecting the accumulator.
        The actual status of a digital output line can also be read.

        TMCL-Mnemonic: GIO <port number>, <bank number>
        """
        cn = NUMBER_COMMANDS['GIO']
        outp = int(port_number)
        bank = int(bank_number)
        if bank == 0:
            if not (0 <= outp <= 3):
                raise TMCLError("GIO: output_number not in range(4) @ bank0")
        elif bank == 1:
            if not (0 <= outp <= 2):
                raise TMCLError("GIO: output_number not in range(3) @ bank1")
        elif bank == 2:
            if not (0 <= outp <= 4):
                raise TMCLError("GIO: output_number not in range(5) @ bank2")
        else:
            raise TMCLError("GIO: bank_number not in range(3)")
        status, value = self._query((0x01, cn, outp, bank, 0x0000))
        if status != STAT_OK:
            raise TMCLError("GIO: got status "+STATUSCODES[status])
        return value

    def sap(self, motor_number, parameter_number, value):
        """
        tmcl_sap(motor_number, parameter_number, value) --> None

        Most of the motion control parameters of the module can be
        specified with the SAP command. The settings will be stored in
        SRAM and therefore are volatile. That is, information will be
        lost after power off. Please use command STAP (store axis
        parameter) in order to store any setting permanently.

        TMCL-Mnemonic: SAP <parameter number>, <motor number>, <value>
        """
        cn = NUMBER_COMMANDS['SAP']
        mn = int(motor_number)
        if not 0 <= mn <= 2:
            raise TMCLError("SAP: motor_number not in range(3)")
        pn, v = self._pn_checkrange(parameter_number, value, "SAP: ")
        status, value = self._query((0x01, cn, pn, mn, v))
        if status != STAT_OK:
            raise TMCLError("SAP: got status "+STATUSCODES[status])
        return None

    def gap(self, motor_number, parameter_number):
        """
        tmcl_gap(self, motor_number, parameter_number) --> int

        Most parameters of the TMCM-1110 can be adjusted individually
        for the axis. With this parameter they can be read out. In
        standalone mode the requested value is also transferred to the
        accumulator register for further processing purposes (such as
        conditioned jumps). In direct mode the value read is only
        output in the value field of the reply (without affecting the
        accumulator).

        TMCL-Mnemonic: GAP <parameter number>, <motor number>
        """
        cn = NUMBER_COMMANDS['GAP']
        mn = int(motor_number)
        pn = int(parameter_number)
        if not 0 <= mn <= 2:
            raise TMCLError("GAP: motor_number not in range(3)")
        if pn not in AXIS_PARAMETER.keys():
            raise TMCLError(prefix+"parameter number not valid")
        status, value = self._query((0x01, cn, pn, mn, 0x0000))
        if status != STAT_OK:
            raise TMCLError("GAP: got status " + STATUSCODES[status]
                                        + ", while querying "+str(pn))
        return value

    def sgp(self, bank_number, parameter_number, value):
        """
        tmcl_sgp(self, bank_number, parameter_number, value) --> None

        Most of the module specific parameters not directly related to
        motion control can be specified and the TMCLTM user variables
        can be changed. Global parameters are related to the host
        interface, peripherals or other application specific variables.
        The different groups of these parameters are organized in banks
        to allow a larger total number for future products. Currently,
        bank 0 and bank 1 are used for global parameters. Bank 2 is
        used for user variables and bank 3 is used for interrupt
        configuration.

        All module settings will automatically be stored non-volatile
        (internal EEPROM of the processor). The TMCLTM user variables
        will not be stored in the EEPROM automatically, but this can
        be done by using STGP commands.

        TMCL-Mnemonic: SGP <parameter number>, <bank number>, <value>
        """
        cn = NUMBER_COMMANDS['SGP']
        bn = int(bank_number)
        pn = int(parameter_number)
        v = int(value)
        if not 0 <= bn <= 3:
            raise TMCLError("SGP: bank_number not in range(4)")
        pn, v = self._pn_checkrange((bn, pn), v)
        status, value = self._query((0x01, cn, pn, bn, v))
        if status != STAT_OK:
            raise TMCLError("SGP: got status "+STATUSCODES[status])
        return None

    def ggp(self, bank_number, parameter_number):
        """
        tmcl_ggp(self, bank_number, parameter_number) --> int

        All global parameters can be read with this function. Global
        parameters are related to the host interface, peripherals or
        application specific variables. The different groups of these
        parameters are organized in banks to allow a larger total
        number for future products. Currently, bank 0 and bank 1 are
        used for global parameters. Bank 2 is used for user variables
        and bank 3 is used for interrupt configuration. Internal
        function: the parameter is read out of the correct position
        in the appropriate device. The parameter format is converted
        adding leading zeros (or ones for negative values).

        TMCL-Mnemonic: GGP <parameter number>, <bank number>
        """
        cn = NUMBER_COMMANDS['GGP']
        bn = int(bank_number)
        pn = int(parameter_number)
        if not 0 <= bn <= 3:
            raise TMCLError("GGP: bank_number not in range(4)")
        if not (bn, pn) in GLOBAL_PARAMETER.keys():
            raise TMCLError("GGP: parameter number not valid")
        status, value = self._query((0x01, cn, pn, bn, 0x0000))
        if status != STAT_OK:
            raise TMCLError("GGP: got status "+STATUSCODES[status])
        return value

    def stap(self, motor_number, parameter_number):
        """
        tmcl_stap(self, motor_number, parameter_number) --> None

        An axis parameter previously set with a Set Axis Parameter
        command (SAP) will be stored permanent. Most parameters are
        automatically restored after power up.

        TMCL-Mnemonic: STAP <parameter number>, <motor number>
        """
        cn = NUMBER_COMMANDS['STAP']
        mn = int(motor_number)
        if not 0 <= mn <= 2:
            raise TMCLError("STAP: motor_number not in range(3)")
        pn = int(parameter_number)
        if not pn in AXIS_PARAMETER.keys():
            raise TMCLError("STAP: parameter number not valid")
        status, value = self._query((0x01, cn, pn, mn, 0x0000))
        if status != STAT_OK:
            raise TMCLError("STAP: got status "+STATUSCODES[status])
        return None

    def rsap(self):
        raise NotImplementedError("yet!")

    def stgp(self):
        raise NotImplementedError("yet!")

    def rsgp(self):
        raise NotImplementedError("yet!")



"""Contains the Device base class."""
import abc

from typing import List, Any, Optional

from mpf.core.machine import MachineController
from mpf.core.logging import LogMixin

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.mode import Mode      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.platform import BasePlatform    # pylint: disable-msg=cyclic-import,unused-import; # noqa


class Device(LogMixin, metaclass=abc.ABCMeta):

    """Generic parent class of for every hardware device in a pinball machine."""

    # String of the config section name
    config_section = None   # type: str

    # String name of the collection
    collection = None       # type: str

    # String of the friendly name of the device class
    class_label = None      # type: str

    # Can a config for this device be empty?
    allow_empty_configs = False

    __slots__ = ["machine", "name", "tags", "platform", "label", "config"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Set up default attributes of every device.

        Args:
        ----
            machine: The machine controller.
            name: Name of the device in config.
        """
        super().__init__()
        self.machine = machine
        self.name = name
        self.tags = []          # type: List[str]
        self.platform = None    # type: Optional[BasePlatform]
        """List of tags applied to this device."""

        self.label = None       # type: Optional[str]
        self.config = dict()    # type: Any
        """Validated dictionary of this device's settings. Note that this will
        map to the YAML-based config specified in the Config Spec section of
        the User Documentation.
        """

    def __lt__(self, other):
        """Compare two devices."""
        return self.name < other.name

    async def device_added_to_mode(self, mode: "Mode") -> None:
        """Add a device to a running mode.

        Args:
        ----
            mode: Mode which loaded the device
        """
        raise AssertionError("Cannot use device {} in mode {}.".format(self.name, mode.name))

    @classmethod
    def get_config_spec(cls):
        """Return config spec for this device."""
        return False

    def load_platform_section(self, platform_section: str):
        """Can be called in _initialize to load the platform section.

        Args:
        ----
            platform_section: Name of the platform section.
        """
        self.platform = self.machine.get_platform_sections(platform_section, self.config['platform'])

    @classmethod
    def prepare_config(cls, config: dict, is_mode_config: bool) -> dict:
        """Return the config prepared for validation.

        Args:
        ----
            config: Config of device
            is_mode_config: Whether this device is loaded in a mode or system-wide

        Returns: Prepared config
        """
        del is_mode_config
        return config

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Return the parsed and validated config.

        Args:
        ----
            config: Config of device
            is_mode_config: Whether this device is loaded in a mode or system-wide
            debug_prefix: Prefix to use when logging.

        Returns: Validated config
        """
        del is_mode_config
        self.machine.config_validator.validate_config(
            self.config_section, config, self.name, "device", prefix=debug_prefix)

        self._configure_device_logging(config)

        return config

    def _configure_device_logging(self, config):

        if config['debug']:
            config['console_log'] = 'full'
            config['file_log'] = 'full'

        self.configure_logging(self.class_label + '.' + self.name,
                               config['console_log'],
                               config['file_log'], url_base=self.class_label)

        self.debug_log("Configuring device with settings: '%s'", config)

    def load_config(self, config: dict):
        """Load config.

        Args:
        ----
            config: Config for device
        """
        self.config = config

        self.tags = self.config['tags']
        self.label = self.config['label']

    def __repr__(self):
        """Return string representation."""
        return '<{self.class_label}.{self.name}>'.format(self=self)

    @classmethod
    def get_config_info(cls):
        """Return config collection and config section.

        Returns (str, str): Tuple with (collection, config_section)
        """
        if not cls.collection or not cls.config_section:
            raise AssertionError("Implement collection and config_section in {}".format(cls))

        return cls.collection, cls.config_section

    async def _initialize(self):
        """Initialise device."""



from unittest.mock import MagicMock,Mock
from mpf.tests.MpfTestCase import MpfTestCase
import mpf.platforms.trinamics_steprocker

class TestTrinamicsStepRocker(MpfTestCase):

    def get_config_file(self):
        return 'trinamics_steprocker.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/trinamics_steprocker/'

    def get_platform(self):
        return False

    def test_rotary(self):
        return
        stepper = self.machine.steppers["velocityStepper"]

        # spin clockwise, 45 degrees per second
        stepper.move_vel_mode( 45 )
        self.assertEqual( 45, stepper._cachedVelocity )

        # stop
        stepper.stop()
        self.assertEqual( 0, stepper._cachedVelocity )

        # spin counter clockwise
        stepper.move_vel_mode( -60 )
        self.assertEqual( -60, stepper._cachedVelocity )

        # stop
        stepper.stop()
        self.assertEqual( 0, stepper._cachedVelocity )

        # try to go too fast
        with self.assertRaises(ValueError):
            stepper.move_vel_mode( stepper.config['velocity_limit'] + 0.01 )

        # try a home / should be rejected
        with self.assertRaises(RuntimeError):
            stepper.home()

        # try a position move / should be rejected
        with self.assertRaises(RuntimeError):
            stepper._move_to_absolute_position(42)

        # try a relative move / should be rejected
        with self.assertRaises(RuntimeError):
            stepper.move_rel_pos( 42 )

    def test_AbsPositionTest(self):
        stepper = self.machine.steppers["positionStepper"]

        # check home/reset
        self.assertEqual(0.0, stepper._current_position)

        # min/max in test file is 0,1 scaling setup for 1.0 = 1 revolution
        stepper._move_to_absolute_position(0.5)
        self.machine.clock.loop.run_until_complete(self.machine.events.wait_for_event('stepper_positionStepper_ready'))
        self.assertEqual(0.5, stepper._current_position)

        # Go to max
        stepper._move_to_absolute_position(1.0)
        self.machine.clock.loop.run_until_complete(self.machine.events.wait_for_event('stepper_positionStepper_ready'))
        self.assertEqual(1.0, stepper._current_position)

        # Go to min
        stepper._move_to_absolute_position(0.0)
        self.machine.clock.loop.run_until_complete(self.machine.events.wait_for_event('stepper_positionStepper_ready'))
        self.assertEqual(0.0, stepper._current_position)

    def setUp(self):
        #self.MockTMCLLib = MagicMock(spec=mpf.platforms.trinamics_steprocker.TMCLDevice)
        self.MockTMCLLib = MagicMock()
        self.MockTMCLLib.rfs.return_value = int(0)
        self.MockTMCLLib.gap.side_effect = self.mock_gap
        self.MockTMCLLib.mvp.side_effect = self.mock_mvp
        self._trinam_pos = 0
        mpf.platforms.trinamics_steprocker.TMCLDevice = MagicMock()
        mpf.platforms.trinamics_steprocker.TMCLDevice.return_value = self.MockTMCLLib
        super().setUp()

    # Mock functionality needed
    def mock_mvp(self, motor_number, cmdtype, value):
        self._trinam_pos = value

    def mock_gap(self, mn, param):
        global _trinam_pos
        if param == 1:
            return self._trinam_pos
        elif param == 8:
            return 1




from mpf.tests.MpfTestCase import MpfTestCase


class TestConfigOldVersion(MpfTestCase):
    def get_config_file(self):
        return 'test_config_interface_missing_version.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/config_interface/'

    def setUp(self):
        self.save_and_prepare_sys_path()

    def tearDown(self):
        self.restore_sys_path()

    def test_config_file_with_old_version(self):
        self.assertRaises(ValueError, super().setUp)
        self.loop.close()



"""Classes which manage BCP transports."""
from collections import defaultdict

from typing import Union

from mpf.core.bcp.bcp_client import BaseBcpClient
from mpf.core.utility_functions import Util

MYPY = False  # noqa
if MYPY:
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class BcpTransportManager:

    """Manages BCP transports."""

    __slots__ = ["_machine", "_transports", "_readers", "_handlers"]

    def __init__(self, machine):
        """Initialise BCP transport manager."""
        self._machine = machine     # type: MachineController
        self._transports = []
        self._readers = {}
        self._handlers = defaultdict(set)
        self._machine.events.add_handler("shutdown", self.shutdown)

    def add_handler_to_transport(self, handler, transport: BaseBcpClient):
        """Register client as handler."""
        if transport is None:
            raise AssertionError("Cannot register None transport.")

        self._handlers[handler].add(transport)

    def remove_transport_from_handle(self, handler, transport: BaseBcpClient):
        """Remove client from a certain handler."""
        if transport in self._handlers[handler]:
            self._handlers[handler].remove(transport)

    def get_transports_for_handler(self, handler):
        """Get clients which registered for a certain handler."""
        return self._handlers.get(handler, [])

    def register_transport(self, transport, future=None, **kwargs):
        """Register a client."""
        del future
        del kwargs
        self._transports.append(transport)
        self._readers[transport] = self._machine.clock.loop.create_task(self._receive_loop(transport))
        self._readers[transport].add_done_callback(Util.raise_exceptions)

    async def _receive_loop(self, transport: BaseBcpClient):
        while True:
            try:
                cmd, kwargs = await transport.read_message()
            except IOError:
                self.unregister_transport(transport)
                return

            await self._machine.bcp.interface.process_bcp_message(cmd, kwargs, transport)

    def unregister_transport(self, transport: BaseBcpClient):
        """Unregister client."""
        if transport in self._transports:
            self._transports.remove(transport)

        # remove transport from all handlers
        for handler in self._handlers:
            if transport in self._handlers[handler]:
                self._handlers[handler].remove(transport)

        if transport in self._readers:
            self._readers[transport].cancel()
            del self._readers[transport]

        if transport.exit_on_close:
            self._machine.stop("BCP client {} disconnected and exit_on_close is set".format(transport.name))

    def get_all_clients(self):
        """Get a list of all clients."""
        return self._transports

    def get_named_client(self, client_name) -> Union[BaseBcpClient, bool]:
        """Get a client by name."""
        for client in self._transports:
            if client.name == client_name:
                return client
        return False

    def send_to_clients(self, clients, bcp_command, **kwargs):
        """Send command to a list of clients."""
        for client in set(clients):
            self.send_to_client(client, bcp_command, **kwargs)

    def send_to_clients_with_handler(self, handler, bcp_command, **kwargs):
        """Send command to clients which registered for a specific handler."""
        clients = self.get_transports_for_handler(handler)
        self.send_to_clients(clients, bcp_command, **kwargs)

    def send_to_client(self, client: BaseBcpClient, bcp_command, **kwargs):
        """Send command to a specific bcp client."""
        try:
            client.send(bcp_command, kwargs)
        except IOError:
            client.stop()
            self.unregister_transport(client)

    def send_to_all_clients(self, bcp_command, **kwargs):
        """Send command to all bcp clients."""
        for client in self._transports:
            self.send_to_client(client, bcp_command, **kwargs)

    def shutdown(self, **kwargs):
        """Prepare the BCP clients for MPF shutdown."""
        del kwargs
        for client in list(self._transports):
            client.stop()
            self.unregister_transport(client)



import unittest
import ruamel.yaml as yaml
from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip

from mpf.file_interfaces.yaml_interface import MpfLoader


class TestYamlInterface(unittest.TestCase):

    def test_round_trip(self):
        self.maxDiff = None

        orig_config = """\
hardware:
    platform: smart_virtual
    driverboards: virtual
    dmd: smartmatrix

config:
-   portconfig.yaml
-   switches.yaml
-   coils.yaml
-   devices.yaml
-   keyboard.yaml
-   virtual.yaml
-   images.yaml

dmd:
    physical: false
    width: 128
    height: 32
    type: color

window:
    elements:
    -   type: virtualdmd
        width: 512
        height: 128
        h_pos: center
        v_pos: center
        pixel_color: ff6600
        dark_color: 220000
        pixel_spacing: 1
    -   type: shape
        shape: box
        width: 516
        height: 132
        color: aaaaaa
        thickness: 2

modes:
-   base
-   airlock_multiball

sound_system:
    buffer: 512
    frequency: 44100
    channels: 1
    initial_volume: 1
    volume_steps: 20
    tracks:
        voice:
            volume: 1
            priority: 2
            simultaneous_sounds: 1
            preload: false
        sfx:
            volume: 1
            priority: 1
            preload: false
            simultaneous_sounds: 3
    stream:
        name: music
        priority: 0
"""
        parsed_config = YamlRoundtrip.process(orig_config)
        saved_config = YamlRoundtrip.save_to_str(parsed_config)

        # print(saved_config)

        self.assertEqual(orig_config, saved_config)

    def test_duplicate_key(self):
        yaml_str = '''

a: 1
b: 2
a: 3


'''
        with self.assertRaises(KeyError):
            yaml.load(yaml_str, Loader=MpfLoader)

    def test_yaml_patches(self):

        # tests our patches to the yaml processor

        config = """

str_1: +1
str_2: 032
str_3: on
str_4: off
str_5: 123e45
str_6: hi
str_7: 2:10
str_8: 2:10.1
bool_1: yes
bool_2: no
bool_3: true
bool_4: false
bool_5: True
bool_6: False
int_1: 123
yaml.scalarfloat.ScalarFloat_1: 1.0

        """

        values = {
            "str_1": "+1",
            "str_2": "032",
            "str_3": "on",
            "str_4": "off",
            "str_5": "123e45",
            "str_6": "hi",
            "str_7": "2:10",
            "str_8": "2:10.1",
            "bool_1": True,
            "bool_2": False,
            "bool_3": True,
            "bool_4": False,
            "bool_5": True,
            "bool_6": False,
            "int_1": 123,
            "yaml.scalarfloat.ScalarFloat_1": 1.0,
        }

        parsed_config = YamlRoundtrip.process(config)

        for k, v in parsed_config.items():
            if not type(v) is eval(k.split('_')[0]):
                raise AssertionError('YAML value "{}" is {}, not {}'.format(v,
                    type(v), eval(k.split('_')[0])))
            self.assertEqual(values[k], v)



from mpf.core.platform import SwitchConfig
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase, MagicMock

from mpf.tests.loop import MockSerial


class BaseMockFast(MockSerial):

    def __init__(self):
        super().__init__()
        self.type = None
        self.queue = []
        self.expected_commands = {}
        self.ignore_commands = {}

    def read(self, length):
        del length
        if not self.queue:
            return
        msg = (self.queue.pop() + '\r').encode()
        return msg

    def read_ready(self):
        return bool(len(self.queue) > 0)

    def write_ready(self):
        return True

    def _parse(self, msg):
        return False

    def write(self, msg):
        """Write message."""
        parts = msg.split(b'\r')
        # remove last newline
        assert parts.pop() == b''

        for part in parts:
            self._handle_msg(part)

        return len(msg)

    def _handle_msg(self, msg):
        msg_len = len(msg)
        cmd = msg.decode()
        # strip newline
        # ignore init garbage
        if cmd == (' ' * 256 * 4):
            return msg_len

        if cmd[:3] == "WD:" and cmd != "WD:1":
            self.queue.append("WD:P")
            return msg_len

        if cmd in self.ignore_commands:
            self.queue.append(cmd[:3] + "P")
            return msg_len

        if self._parse(cmd):
            return msg_len

        if cmd in self.expected_commands:
            if self.expected_commands[cmd]:
                self.queue.append(self.expected_commands[cmd])
            del self.expected_commands[cmd]
            return msg_len
        else:
            raise Exception(self.type + ": " + str(cmd))

    def stop(self):
        pass


class MockFastDmd(BaseMockFast):
    def __init__(self):
        super().__init__()
        self.type = "DMD"

    def write(self, msg):
        """Write message."""
        parts = msg.split(b'\r')

        # remove last newline
        if parts[len(parts) - 1] == b'':
            parts.pop() == b''

        for part in parts:
            self._handle_msg(part)

        return len(msg)

    def _handle_msg(self, msg):
        msg_len = len(msg)
        if msg == (b' ' * 256 * 4):
            return msg_len

        cmd = msg

        if cmd[:3] == "WD:":
            self.queue.append("WD:P")
            return msg_len

        if cmd in self.ignore_commands:
            self.queue.append(cmd[:3] + "P")
            return msg_len

        if cmd in self.expected_commands:
            if self.expected_commands[cmd]:
                self.queue.append(self.expected_commands[cmd])
            del self.expected_commands[cmd]
            return msg_len
        else:
            raise Exception(self.type + ": " + str(cmd))


class MockFastRgb(BaseMockFast):
    def __init__(self):
        super().__init__()
        self.type = "RGB"
        self.ignore_commands["L1:23,FF"] = True
        self.leds = {}

    def _parse(self, cmd):
        if cmd[:3] == "RS:":
            remaining = cmd[3:]
            while True:
                self.leds[remaining[0:2]] = remaining[2:8]
                remaining = remaining[9:]

                if not remaining:
                    break

            self.queue.append("RX:P")
            return True

class MockFastNet(BaseMockFast):
    def __init__(self):
        super().__init__()
        self.type = "NET"


class TestFast(MpfTestCase):
    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/fast/'

    def get_platform(self):
        return False

    def _mock_loop(self):
        self.clock.mock_serial("com4", self.net_cpu)
        self.clock.mock_serial("com5", self.rgb_cpu)
        self.clock.mock_serial("com6", self.dmd_cpu)

    def tearDown(self):
        self.dmd_cpu.expected_commands = {
            b'BL:AA55': "!SRE"
        }
        self.rgb_cpu.expected_commands = {
            "BL:AA55": "!SRE"
        }
        self.net_cpu.expected_commands = {
            "WD:1": "WD:P"
        }
        super().tearDown()
        self.assertFalse(self.net_cpu.expected_commands)
        self.assertFalse(self.rgb_cpu.expected_commands)
        self.assertFalse(self.dmd_cpu.expected_commands)

    def setUp(self):
        self.expected_duration = 2
        self.net_cpu = MockFastNet()
        self.rgb_cpu = MockFastRgb()
        self.dmd_cpu = MockFastDmd()

        self.dmd_cpu.expected_commands = {
            b'ID:': 'ID:DMD FP-CPU-002-1 00.88',

        }
        self.rgb_cpu.expected_commands = {
            'ID:': 'ID:RGB FP-CPU-002-1 00.89',
            "RF:0": "RF:P",
            "RA:000000": "RA:P",
            "RF:00": "RF:P",
        }
        self.net_cpu.expected_commands = {
            'BR:': '#!B:02',    # there might be some garbage in front of the command
            'ID:': 'ID:NET FP-CPU-002-1 01.03',
            'NN:00': 'NN:00,FP-I/O-3208-2   ,01.00,08,20,04,06,00,00,00,00',     # 3208 board
            'NN:01': 'NN:01,FP-I/O-0804-1   ,01.00,04,08,04,06,00,00,00,00',     # 0804 board
            'NN:02': 'NN:02,FP-I/O-1616-2   ,01.00,10,10,04,06,00,00,00,00',     # 1616 board
            'NN:03': 'NN:03,FP-I/O-1616-2   ,01.00,10,10,04,06,00,00,00,00',     # 1616 board
            'NN:04': 'NN:04,,,,,,,,,,',     # no board
            "SA:": "SA:01,00,09,050000000000000000",
            "SN:01,01,04,04": "SN:P",
            "SN:02,01,04,04": "SN:P",
            "SN:03,01,04,04": "SN:P",
            "SN:0B,01,04,04": "SN:P",
            "SN:0C,01,04,04": "SN:P",
            "SN:16,01,04,04": "SN:P",
            "SN:07,01,1A,05": "SN:P",
            "SN:1A,01,04,04": "SN:P",
            "SN:39,01,04,04": "SN:P",
            "DN:01,00,00,00": "DN:P",
            "DN:04,00,00,00": "DN:P",
            "DN:06,00,00,00": "DN:P",
            "DN:07,00,00,00": "DN:P",
            "DN:11,00,00,00": "DN:P",
            "DN:12,00,00,00": "DN:P",
            "DN:13,00,00,00": "DN:P",
            "DN:20,00,00,00": "DN:P",
            "DN:21,00,00,00": "DN:P",
            "DN:01,C1,00,18,00,FF,FF,00": "DN:P",   # configure digital output
            "XO:03,7F": "XO:P",
            "XO:14,7F": "XO:P"
        }

        super().setUp()
        self.advance_time_and_run()
        self.assertFalse(self.net_cpu.expected_commands)
        self.assertFalse(self.rgb_cpu.expected_commands)
        self.assertFalse(self.dmd_cpu.expected_commands)

        # test io board detection
        self.assertEqual(4, len(self.machine.default_platform.io_boards))
        self.assertEqual(32, self.machine.default_platform.io_boards[0].switch_count)
        self.assertEqual(8, self.machine.default_platform.io_boards[0].driver_count)
        self.assertEqual(8, self.machine.default_platform.io_boards[1].switch_count)
        self.assertEqual(4, self.machine.default_platform.io_boards[1].driver_count)
        self.assertEqual(16, self.machine.default_platform.io_boards[2].switch_count)
        self.assertEqual(16, self.machine.default_platform.io_boards[2].driver_count)
        self.assertEqual(16, self.machine.default_platform.io_boards[3].switch_count)
        self.assertEqual(16, self.machine.default_platform.io_boards[3].driver_count)

        self.assertEqual("00.88", self.machine.variables.get_machine_var("fast_dmd_firmware"))
        self.assertEqual("FP-CPU-002-1", self.machine.variables.get_machine_var("fast_dmd_model"))
        self.assertEqual("00.89", self.machine.variables.get_machine_var("fast_rgb_firmware"))
        self.assertEqual("FP-CPU-002-1", self.machine.variables.get_machine_var("fast_rgb_model"))
        self.assertEqual("01.03", self.machine.variables.get_machine_var("fast_net_firmware"))
        self.assertEqual("FP-CPU-002-1", self.machine.variables.get_machine_var("fast_net_model"))

    def test_coils(self):
        self._test_pulse()
        self._test_long_pulse()
        self._test_enable_exception()
        self._test_allow_enable()
        self._test_pwm_ssm()
        self._test_coil_configure()

        # test hardware scan
        info_str = """NET CPU: NET FP-CPU-002-1 01.03
RGB CPU: RGB FP-CPU-002-1 00.89
DMD CPU: DMD FP-CPU-002-1 00.88

Boards:
Board 0 - Model: FP-I/O-3208-2    Firmware: 01.00 Switches: 32 Drivers: 8
Board 1 - Model: FP-I/O-0804-1    Firmware: 01.00 Switches: 8 Drivers: 4
Board 2 - Model: FP-I/O-1616-2    Firmware: 01.00 Switches: 16 Drivers: 16
Board 3 - Model: FP-I/O-1616-2    Firmware: 01.00 Switches: 16 Drivers: 16
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_coil_configure(self):
        self.assertEqual("FAST Board 0", self.machine.coils["c_test"].hw_driver.get_board_name())
        self.assertEqual("FAST Board 3", self.machine.coils["c_flipper_hold"].hw_driver.get_board_name())
        # last driver on board
        self.net_cpu.expected_commands = {
            "DN:2B,00,00,00": "DN:P"
        }
        coil = self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '3-15',
                                                              {"connection": "network", "recycle_ms": 10})
        self.assertEqual('2B', coil.number)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # board 0 has 8 drivers. configuring driver 9 should not work
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '0-8',
                                                           {"connection": "network", "recycle_ms": 10})

        # only boards 0-3 exist
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '4-0',
                                                           {"connection": "network", "recycle_ms": 10})

        # only 8 + 4 + 16 + 16 = 44 = 0x2C driver exist
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_driver(self.machine.coils["c_test"].hw_driver.config, '44',
                                                           {"connection": "network", "recycle_ms": 10})

    def _test_pulse(self):
        self.net_cpu.expected_commands = {
            "DN:04,81,00,10,17,FF,00,00,00": "DN:P",
            "TN:04,01": "TN:P"
        }
        # pulse coil 4
        self.machine.coils["c_test"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_long_pulse(self):
        # enable command
        self.net_cpu.expected_commands = {
            "DN:12,C1,00,18,00,FF,FF,00": "DN:P"
        }
        self.machine.coils["c_long_pulse"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # disable command
        self.net_cpu.expected_commands = {
            "TN:12,02": "TN:P"
        }

        self.advance_time_and_run(1)
        # pulse_ms is 2000ms, so after 1s, this should not be sent
        self.assertTrue(self.net_cpu.expected_commands)

        self.advance_time_and_run(1)
        # but after 2s, it should be
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()
            self.advance_time_and_run(.1)

    def _test_allow_enable(self):
        self.net_cpu.expected_commands = {
            "DN:06,C1,00,18,17,FF,FF,00": "DN:P"
        }
        self.machine.coils["c_test_allow_enable"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_pwm_ssm(self):
        self.net_cpu.expected_commands = {
            "DN:13,C1,00,18,0A,FF,84224244,00": "DN:P"
        }
        self.machine.coils["c_hold_ssm"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_nano_reboot(self):
        # NANO reboots
        self.net_cpu.queue.append("!B:00")
        self.advance_time_and_run(.1)
        # assert that MPF will stop
        self.assertTrue(self.machine.stop_future.done())

    def test_rules(self):
        self._test_enable_exception_hw_rule()
        self._test_two_rules_one_switch()
        self._test_hw_rule_pulse()
        self._test_hw_rule_pulse_pwm32()
        self._test_hw_rule_pulse_inverted_switch()
        self._test_hw_rule_same_board()

    def _test_hw_rule_same_board(self):
        self.net_cpu.expected_commands = {
            "DN:21,01,07,10,0A,FF,00,00,14": "DN:P"
        }
        # coil and switch are on different boards but first 8 switches always work
        self.machine.autofires["ac_different_boards"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # switch and coil on board 3. should work
        self.net_cpu.expected_commands = {
            "DN:21,01,39,10,0A,FF,00,00,14": "DN:P",
            "SN:39,01,02,02": "SN:P"
        }
        self.machine.autofires["ac_board_3"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "DN:10,01,03,10,0A,89,00,00,14": "DN:P",
        }
        # coil and switch are on different boards
        with self.assertRaises(AssertionError):
            self.machine.autofires["ac_broken_combination"].enable()
            self.advance_time_and_run(.1)

    def _test_enable_exception_hw_rule(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = 1.0
            self.machine.flippers["f_test_single"].enable()

        self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = None

    def _test_two_rules_one_switch(self):
        self.net_cpu.expected_commands = {
            "SN:03,01,02,02": "SN:P",
            "DN:04,01,03,10,17,FF,00,00,1B": "DN:P",
            "DN:06,01,03,10,17,FF,00,00,2E": "DN:P"
        }
        self.post_event("ac_same_switch")
        self.hit_and_release_switch("s_flipper")
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse(self):
        self.net_cpu.expected_commands = {
            "DN:07,01,16,10,0A,FF,00,00,14": "DN:P",  # hw rule
            "SN:16,01,02,02": "SN:P"                  # debounce quick on switch
        }
        self.machine.autofires["ac_slingshot_test"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "DN:07,81": "DN:P"
        }
        self.machine.autofires["ac_slingshot_test"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse_pwm32(self):
        self.net_cpu.expected_commands = {
            "DN:11,81,00,10,0A,AAAAAAAA,00,00,00": "DN:P",
            "TN:11,01": "TN:P"
        }
        self.machine.coils["c_pulse_pwm32_mask"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "DN:11,C1,00,18,0A,AAAAAAAA,4A4A4A4A,00": "DN:P"
        }
        self.machine.coils["c_pulse_pwm32_mask"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse_inverted_switch(self):
        self.net_cpu.expected_commands = {
            "DN:07,11,1A,10,0A,FF,00,00,14": "DN:P",
            "SN:1A,01,02,02": "SN:P"
        }
        self.machine.autofires["ac_inverted_switch"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_firmware_update(self):
        commands = []

        def _catch_update(cmd):
            commands.append(cmd)
            return len(cmd)
        parse_func = self.net_cpu.write
        self.net_cpu.write = _catch_update
        output = self.machine.default_platform.update_firmware()
        self.advance_time_and_run()
        self.net_cpu.write = parse_func
        # check if we send the dummy update
        self.assertEqual([b'BL:AA55\r>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                          b'>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                          b'>>>>>>>>>>>>>>>>>>>>>>>>>\rBL:AA55\r<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<'
                          b'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<'
                          b'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\rBL:AA55\r>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                          b'>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                          b'>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\rDUMMY UPDAT'
                          b'E\r', b'WD:3e8\r', b'WD:3e8\r'], commands)
        expected_output = """NET CPU is version 01.03
Found an update to version 1.04 for the NET CPU. Will flash file firmware/FAST_NET_01_04_00.txt
Update done.
"""
        self.assertEqual(expected_output, output)

    def test_servo(self):
        # go to min position
        self.net_cpu.expected_commands = {
                "XO:03,00": "XO:P"
        }
        self.machine.servos["servo1"].go_to_position(0)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # go to max position
        self.net_cpu.expected_commands = {
                "XO:03,FF": "XO:P"
        }
        self.machine.servos["servo1"].go_to_position(1)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _switch_hit_cb(self, **kwargs):
        self.switch_hit = True

    def test_switches(self):
        self._test_switch_changes()
        self._test_switch_changes_nc()
        self._test_switch_configure()

    def _test_switch_configure(self):
        # last switch on first board
        self.net_cpu.expected_commands = {
            "SN:1F,01,04,04": "SN:P"
        }
        self.machine.default_platform.configure_switch('0-31', SwitchConfig(name="", debounce='auto', invert=0), {})
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # next should not work
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('0-32', SwitchConfig(name="", debounce='auto', invert=0), {})

        self.net_cpu.expected_commands = {
            "SN:47,01,04,04": "SN:P"
        }
        self.machine.default_platform.configure_switch('3-15', SwitchConfig(name="", debounce='auto', invert=0), {})
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # invalid board
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('4-0', SwitchConfig(name="", debounce='auto', invert=0), {})

        # last switch is 0x47. 0x48 = 72
        with self.assertRaises(AssertionError):
            self.machine.default_platform.configure_switch('72', SwitchConfig(name="", debounce='auto', invert=0), {})

    def _test_switch_changes(self):
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_flipper_eos", 1)

        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test", 0)
        self.assertFalse(self.switch_hit)

        self.machine.events.add_handler("s_test_active", self._switch_hit_cb)
        self.machine.default_platform.process_received_message("-N:07", "NET")
        self.advance_time_and_run(1)

        self.assertTrue(self.switch_hit)
        self.assertSwitchState("s_test", 1)
        self.switch_hit = False

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 1)

        self.machine.default_platform.process_received_message("/N:07", "NET")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 0)

    def _test_switch_changes_nc(self):
        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test_nc", 1)
        self.assertFalse(self.switch_hit)

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 1)

        self.machine.default_platform.process_received_message("-N:1A", "NET")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 0)

        self.machine.events.add_handler("s_test_nc_active", self._switch_hit_cb)
        self.machine.default_platform.process_received_message("/N:1A", "NET")
        self.advance_time_and_run(1)

        self.assertSwitchState("s_test_nc", 1)
        self.assertTrue(self.switch_hit)
        self.switch_hit = False

    def test_flipper_single_coil(self):
        # manual flip no hw rule
        self.net_cpu.expected_commands = {
            "DN:20,81,00,10,0A,FF,00,00,00": "DN:P",
            "TN:20,01": "TN:P",
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable no hw rule
        self.net_cpu.expected_commands = {
            "DN:20,C1,00,18,0A,FF,01,00": "DN:P"
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable no hw rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P"
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # flipper rule enable
        self.net_cpu.expected_commands = {
            "DN:20,01,01,18,0B,FF,01,00,00": "DN:P",
            "SN:01,01,02,02": "SN:P"
        }
        self.machine.flippers["f_test_single"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip with hw rule in action
        self.net_cpu.expected_commands = {
            "DN:20,81,00,10,0A,FF,00,00,00": "DN:P",    # configure pulse
            "TN:20,01": "TN:P",                         # pulse
            "DN:20,01,01,18,0B,FF,01,00,00": "DN:P",    # restore rule
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip with hw rule in action without reconfigure (same pulse)
        self.net_cpu.expected_commands = {
            "TN:20,01": "TN:P",                         # pulse
        }
        self.machine.coils["c_flipper_main"].pulse(11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable with hw rule (same pulse)
        self.net_cpu.expected_commands = {
            "TN:20,03": "TN:P"
        }
        self.machine.coils["c_flipper_main"].enable(pulse_ms=11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable with hw rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P",
            "TN:20,00": "TN:P"   # reenable autofire rule
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable with hw rule (different pulse)
        self.net_cpu.expected_commands = {
            "DN:20,C1,00,18,0A,FF,01,00": "DN:P",       # configure pwm + enable
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable with hw rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P",
            "DN:20,01,01,18,0B,FF,01,00,00": "DN_P",    # configure rules
            "TN:20,00": "TN:P"                          # reenable autofire rule
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # disable rule
        self.net_cpu.expected_commands = {
            "DN:20,81": "DN:P"
        }
        self.machine.flippers["f_test_single"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip no hw rule
        self.net_cpu.expected_commands = {
            "DN:20,81,00,10,0A,FF,00,00,00": "DN:P",
            "TN:20,01": "TN:P",
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip again with cached config
        self.net_cpu.expected_commands = {
            "TN:20,01": "TN:P",
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_flipper_two_coils(self):
        # we pulse the main coil (20)
        # hold coil (21) is pulsed + enabled
        self.net_cpu.expected_commands = {
            "DN:20,01,01,18,0A,FF,00,00,00": "DN:P",
            "DN:21,01,01,18,0A,FF,01,00,00": "DN:P",
            "SN:01,01,02,02": "SN:P",
        }
        self.machine.flippers["f_test_hold"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "DN:20,81": "DN:P",
            "DN:21,81": "DN:P"
        }
        self.machine.flippers["f_test_hold"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_dmd_update(self):

        # test configure
        dmd = self.machine.default_platform.configure_dmd()

        # test set frame to buffer
        frame = bytearray()
        for i in range(4096):
            frame.append(64 + i % 192)

        frame = bytes(frame)

        # test draw
        self.dmd_cpu.expected_commands = {
            b'BM:' + frame: False
        }
        dmd.update(frame)

        self.advance_time_and_run(.1)

        self.assertFalse(self.dmd_cpu.expected_commands)

    def test_bootloader_crash(self):
        # Test that the machine stops if the RGB processor sends a bootloader msg
        self.machine.stop = MagicMock()
        self.machine.default_platform.process_received_message("!B:00", "RGB")
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.stop.called)

    def test_bootloader_crash_ignored(self):
        # Test that RGB processor bootloader msgs can be ignored
        self.machine.default_platform.config['ignore_rgb_crash'] = True
        self.mock_event('fast_rgb_rebooted')
        self.machine.stop = MagicMock()
        self.machine.default_platform.process_received_message("!B:00", "RGB")
        self.advance_time_and_run(1)
        self.assertFalse(self.machine.stop.called)
        self.assertEventCalled('fast_rgb_rebooted')

    def test_lights_and_leds(self):
        self._test_matrix_light()
        self._test_pdb_gi_light()
        self._test_pdb_led()

    def _test_matrix_light(self):
        # test enable of matrix light
        self.net_cpu.expected_commands = {
            "L1:23,FF": "L1:P",
        }
        self.machine.lights["test_pdb_light"].on()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # test enable of matrix light with brightness
        self.net_cpu.expected_commands = {
            "L1:23,80": "L1:P",
        }
        self.machine.lights["test_pdb_light"].on(brightness=128)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # test disable of matrix light
        self.net_cpu.expected_commands = {
            "L1:23,00": "L1:P",
        }
        self.machine.lights["test_pdb_light"].off()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # test disable of matrix light with brightness
        self.net_cpu.expected_commands = {
            "L1:23,00": "L1:P",
        }
        self.machine.lights["test_pdb_light"].on(brightness=255, fade_ms=100)
        self.advance_time_and_run(.02)
        self.assertFalse(self.net_cpu.expected_commands)

        # step 1
        self.net_cpu.expected_commands = {
            "L1:23,32": "L1:P",
            "L1:23,33": "L1:P",
        }
        self.advance_time_and_run(.02)
        self.assertEqual(1, len(self.net_cpu.expected_commands))

        # step 2
        self.net_cpu.expected_commands = {
            "L1:23,65": "L1:P",
            "L1:23,66": "L1:P",
        }
        self.advance_time_and_run(.02)
        self.assertEqual(1, len(self.net_cpu.expected_commands))

        # step 3
        self.net_cpu.expected_commands = {
            "L1:23,98": "L1:P",
            "L1:23,99": "L1:P",
        }
        self.advance_time_and_run(.02)
        self.assertEqual(1, len(self.net_cpu.expected_commands))

        # step 4
        self.net_cpu.expected_commands = {
            "L1:23,CB": "L1:P",
            "L1:23,CC": "L1:P",
        }
        self.advance_time_and_run(.02)
        self.assertEqual(1, len(self.net_cpu.expected_commands))

        # step 5
        self.net_cpu.expected_commands = {
            "L1:23,FE": "L1:P",
            "L1:23,FF": "L1:P",
        }
        self.advance_time_and_run(.02)
        self.assertEqual(1, len(self.net_cpu.expected_commands))

        # step 6 if step 5 did not send FF
        if "L1:23,FE" not in self.net_cpu.expected_commands:
            self.net_cpu.expected_commands = {
                "L1:23,FF": "L1:P",
            }
            self.advance_time_and_run(.02)
            self.assertFalse(self.net_cpu.expected_commands)

    def _test_pdb_gi_light(self):
        # test gi on
        device = self.machine.lights["test_gi"]
        self.net_cpu.expected_commands = {
            "GI:2A,FF": "GI:P",
        }
        device.on()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "GI:2A,80": "GI:P",
        }
        device.on(brightness=128)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "GI:2A,F5": "GI:P",
        }
        device.on(brightness=245)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # test gi off
        self.net_cpu.expected_commands = {
            "GI:2A,00": "GI:P",
        }
        device.off()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "GI:2A,F5": "GI:P",
        }
        device.on(brightness=245)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "GI:2A,00": "GI:P",
        }
        device.on(brightness=0)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_pdb_led(self):
        self.advance_time_and_run()
        device = self.machine.lights["test_led"]
        device2 = self.machine.lights["test_led2"]
        self.assertEqual("000000", self.rgb_cpu.leds['97'])
        self.assertEqual("000000", self.rgb_cpu.leds['98'])
        # test led on
        device.on()
        self.advance_time_and_run(1)
        self.assertEqual("ffffff", self.rgb_cpu.leds['97'])
        self.assertEqual("000000", self.rgb_cpu.leds['98'])

        device2.color("001122")

        # test led off
        device.off()
        self.advance_time_and_run(1)
        self.assertEqual("000000", self.rgb_cpu.leds['97'])
        self.assertEqual("001122", self.rgb_cpu.leds['98'])

        # test led color
        device.color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(1)
        self.assertEqual("02172a", self.rgb_cpu.leds['97'])



from mpf.tests.MpfTestCase import MpfTestCase

class TestSwitchPlayer(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/switch_player/'

    def setUp(self):
        self.machine_config_patches['mpf']['plugins'] = ['mpf.plugins.switch_player.SwitchPlayer']
        super().setUp()

    def _sw_handler(self):
        self.hits += 1

    def test_switch_player(self):
        self.hits = 0
        self.machine.switch_controller.add_switch_handler("s_test3", self._sw_handler)
        self.post_event("test_start")
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(0, self.hits)
        self.advance_time_and_run(0.1)
        self.assertSwitchState("s_test1", 1)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.advance_time_and_run(0.6)
        self.assertSwitchState("s_test1", 1)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(1, self.hits)
        self.advance_time_and_run(0.1)
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(1, self.hits)
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 1)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(1, self.hits)
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 1)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(2, self.hits)
        self.advance_time_and_run(0.1)
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(2, self.hits)
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test1", 0)
        self.assertSwitchState("s_test2", 0)
        self.assertSwitchState("s_test3", 0)
        self.assertEqual(3, self.hits)




from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceNoPlungerSwitch(MpfTestCase):

    def get_config_file(self):
        return 'test_ball_device_no_plunger_switch.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def put_ball_in_trough(self):
        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.advance_time_and_run(1)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 1)

        self.trough_coil = self.machine.coils["trough_eject"]
        self.trough_coil.pulse = MagicMock()

    def test_add_ball_to_pf(self):
        self.put_ball_in_trough()

        # Request ball to pf which puts it in the plunger lane
        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(.1)

        self.assertEqual(1, self.trough_coil.pulse.called)

        self.machine.switch_controller.process_switch('s_trough_1', 0)

        self.advance_time_and_run(11)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)

        # hit some playfield switches
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)

        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)

        # drain
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 1)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 0)



"""Mode timers."""
from typing import List, Optional

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.delays import DelayManager
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player
from mpf.core.mode import Mode

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.clock import PeriodicTask     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.events import EventHandlerKey     # pylint: disable-msg=cyclic-import,unused-import


# pylint: disable-msg=too-many-instance-attributes
@DeviceMonitor("running", "ticks", "end_value", "max_value", "start_value")
class Timer(ModeDevice):

    """Parent class for a mode timer.

    Args:
    ----
        machine: The main MPF MachineController object.
        name: The string name of this timer.
    """

    config_section = 'timers'
    collection = 'timers'
    class_label = 'timer'

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise mode timer."""
        super().__init__(machine, name)
        self.machine = machine
        self.name = name

        self.running = False
        self.start_value = None             # type: Optional[int]
        self.restart_on_complete = None     # type: Optional[bool]
        self._ticks = 0
        self.tick_var = None                # type: Optional[str]
        self.tick_secs = None               # type: Optional[float]
        self.player = None                  # type: Optional[Player]
        self.end_value = None               # type: Optional[int]
        self.max_value = None               # type: Optional[int]
        self.ticks_remaining = None         # type: Optional[int]
        self.direction = None               # type: Optional[str]
        self.timer = None                   # type: Optional[PeriodicTask]
        self.event_keys = list()            # type: List[EventHandlerKey]
        self.delay = None                   # type: Optional[DelayManager]

    async def device_added_to_mode(self, mode: Mode) -> None:
        """Device added in mode."""
        await super().device_added_to_mode(mode)
        self.tick_var = '{}_{}_tick'.format(mode.name, self.name)

    async def _initialize(self):
        await super()._initialize()
        self.ticks_remaining = 0
        self.max_value = self.config['max_value']
        self.direction = self.config['direction']
        self.tick_secs = None
        self.timer = None
        self.event_keys = list()
        self.delay = DelayManager(self.machine)

        self.restart_on_complete = self.config['restart_on_complete']
        self.end_value = None
        self.start_value = None
        self.ticks = None

        if self.config['debug']:
            self.configure_logging('Timer.' + self.name,
                                   'full', 'full')
        else:
            self.configure_logging('Timer.' + self.name,
                                   self.config['console_log'],
                                   self.config['file_log'])

        self.debug_log("----------- Initial Values -----------")
        self.debug_log("running: %s", self.running)
        self.debug_log("start_value: %s", self.start_value)
        self.debug_log("restart_on_complete: %s", self.restart_on_complete)
        self.debug_log("_ticks: %s", self.ticks)
        self.debug_log("end_value: %s", self.end_value)
        self.debug_log("ticks_remaining: %s", self.ticks_remaining)
        self.debug_log("max_value: %s", self.max_value)
        self.debug_log("direction: %s", self.direction)
        self.debug_log("tick_secs: %s", self.tick_secs)
        self.debug_log("--------------------------------------")

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Set up control events when mode is loaded."""
        del mode
        self.tick_secs = self.config['tick_interval'].evaluate([])

        try:
            self.end_value = self.config['end_value'].evaluate([])
        except AttributeError:
            self.end_value = None

        if self.direction == 'down' and not self.end_value:
            self.end_value = 0  # need it to be 0 not None

        self.start_value = self.config['start_value'].evaluate([])
        self.ticks = self.start_value

        self.player = player
        if self.config['control_events']:
            self._setup_control_events(self.config['control_events'])

        if self.config['start_running']:
            self.start()

    @property
    def ticks(self):
        """Return ticks."""
        return self._ticks

    @ticks.setter
    def ticks(self, value):
        self._ticks = value

        try:
            self.player[self.tick_var] = value
            '''player_var: (mode)_(timer)_tick

            desc: Stores the current tick value for the timer from the mode
            (mode) with the time name (timer). For example, a timer called
            "my_timer" which is in the config for "mode1" will store its tick
            value in the player variable ``mode1_my_timer_tick``.
            '''

        except TypeError:
            pass

    @property
    def can_exist_outside_of_game(self):
        """Timer can live outside of games."""
        return True

    def _setup_control_events(self, event_list):
        self.debug_log("Setting up control events")

        kwargs = {}
        for entry in event_list:
            if entry['action'] in ('add', 'subtract', 'jump', 'pause', 'set_tick_interval'):
                handler = getattr(self, entry['action'])
                kwargs = {'timer_value': entry['value']}

            elif entry['action'] in ('start', 'stop', 'reset', 'restart'):
                handler = getattr(self, entry['action'])

            elif entry['action'] == 'change_tick_interval':
                handler = self.change_tick_interval
                kwargs = {'change': entry['value']}

            elif entry['action'] == 'reset_tick_interval':
                handler = self.set_tick_interval
                kwargs = {'timer_value': self.config['tick_interval'].evaluate([])}

            else:
                raise AssertionError("Invalid control_event action {} in mode".
                                     format(entry['action']), self.name)

            self.event_keys.append(
                self.machine.events.add_handler(entry['event'], handler, **kwargs))

    def _remove_control_events(self):
        self.debug_log("Removing control events")

        for key in self.event_keys:
            self.machine.events.remove_handler_by_key(key)

    def reset(self, **kwargs):
        """Reset this timer based to the starting value that's already been configured.

        Does not start or stop the timer.

        Args:
        ----
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.debug_log("Resetting timer. New value: %s", self.start_value)

        self.jump(self.start_value)

    def start(self, **kwargs):
        """Start this timer based on the starting value that's already been configured.

        Use jump() if you want to set the starting time value.

        Args:
        ----
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        # do not start if timer is already running
        if self.running:
            return

        self.info_log("Starting Timer.")

        if self._check_for_done():
            return

        self.running = True

        self.delay.remove('pause')
        self._create_system_timer()

        self.machine.events.post('timer_' + self.name + '_started',
                                 ticks=self.ticks,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_started

        desc: The timer named (name) has just started.

        args:
            ticks: The current tick number this timer is at.
            ticks_remaining: The number of ticks in this timer remaining.
        '''

        self._post_tick_events()
        # since lots of slides and stuff are tied to the timer tick, we want
        # to post an initial tick event also that represents the starting
        # timer value.

    def restart(self, **kwargs):
        """Restart the timer by resetting it and then starting it.

        Essentially this is just a reset() then a start().

        Args:
        ----
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs
        self.reset()
        self.start()

    def stop(self, **kwargs):
        """Stop the timer and posts the 'timer_<name>_stopped' event.

        Args:
        ----
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.info_log("Stopping Timer")

        self.delay.remove('pause')

        self.running = False
        self._remove_system_timer()

        self.machine.events.post('timer_' + self.name + '_stopped',
                                 ticks=self.ticks,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_stopped

        desc: The timer named (name) has stopped.

        This event is posted any time the timer stops, whether it stops because
        it ended or because it was stopped early by some other event.

        args:
            ticks: The current tick number this timer is at.
            ticks_remaining: The number of ticks in this timer remaining.
        '''

    def pause(self, timer_value=0, **kwargs):
        """Pause the timer and posts the 'timer_<name>_paused' event.

        Args:
        ----
            timer_value: How many seconds you want to pause the timer for. Note
                that this pause time is real-world seconds and does not take
                into consideration this timer's tick interval.
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        if not timer_value:
            pause_ms = 0  # make sure it's not None, etc.
        else:
            pause_ms = self._get_timer_value(timer_value) * 1000  # delays happen in ms

        self.info_log("Pausing Timer for %s ms", pause_ms)

        self.running = False

        self._remove_system_timer()
        self.machine.events.post('timer_' + self.name + '_paused',
                                 ticks=self.ticks,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_paused

        desc: The timer named (name) has paused.

        args:
            ticks: The current tick number this timer is at.
            ticks_remaining: The number of ticks in this timer remaining.
        '''

        if pause_ms > 0:
            self.delay.add(name='pause', ms=pause_ms, callback=self.start)

    def timer_complete(self, **kwargs):
        """Automatically called when this timer completes.

        Posts the 'timer_<name>_complete' event. Can be manually called to mark
        this timer as complete.

        Args:
        ----
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.info_log("Timer Complete")

        self.stop()

        self.machine.events.post('timer_' + self.name + '_complete',
                                 ticks=self.ticks,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_complete

        desc: The timer named (name) has completed.

        Note that this timer may reset and start again after this event is
        posted, depending on its settings.

        args:
            ticks: The current tick number this timer is at.
            ticks_remaining: The number of ticks in this timer remaining.
        '''

        if self.restart_on_complete:

            self.debug_log("Restart on complete: True")

            self.reset()
            self.start()

    def _timer_tick(self):
        # Automatically called by the core clock each tick
        self.debug_log("Timer Tick")

        if not self.running:
            self.debug_log("Timer is not running. Will remove.")

            self._remove_system_timer()
            return

        if self.direction == 'down':
            self.ticks -= 1
        else:
            self.ticks += 1

        self._post_tick_events()

    def _post_tick_events(self):

        if not self._check_for_done():
            self.machine.events.post('timer_' + self.name + '_tick',
                                     ticks=self.ticks,
                                     ticks_remaining=self.ticks_remaining)
            '''event: timer_(name)_tick

            desc: The timer named (name) has just counted down (or up,
            depending on its settings).

            args:
                ticks: The new tick number this timer is at.
                ticks_remaining: The new number of ticks in this timer
                    remaining.
            '''

            self.debug_log("Ticks: %s, Remaining: %s",
                           self.ticks,
                           self.ticks_remaining)

    def add(self, timer_value, **kwargs):
        """Add ticks to this timer.

        Args:
        ----
            timer_value: The number of ticks you want to add to this timer's
                current value.
            kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        timer_value = self._get_timer_value(timer_value)
        ticks_added = timer_value

        new_value = self.ticks + ticks_added

        if self.max_value and new_value > self.max_value:
            new_value = self.max_value

        self.ticks = new_value
        ticks_added = new_value - timer_value

        self.machine.events.post('timer_' + self.name + '_time_added',
                                 ticks=self.ticks,
                                 ticks_added=ticks_added,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_time_added

        desc: The timer named (name) has just had time added to it.

        args:
            ticks: The new tick number this timer is at.
            ticks_remaining: The new number of ticks in this timer remaining.
            ticks_added: How many ticks were just added.
        '''

        self._check_for_done()

    def subtract(self, timer_value, **kwargs):
        """Subtract ticks from this timer.

        Args:
        ----
            timer_value: The number of ticks you want to subtract from this
                timer's current value.
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        ticks_subtracted = self._get_timer_value(timer_value)

        self.ticks -= ticks_subtracted

        self.machine.events.post('timer_' + self.name + '_time_subtracted',
                                 ticks=self.ticks,
                                 ticks_subtracted=ticks_subtracted,
                                 ticks_remaining=self.ticks_remaining)
        '''event: timer_(name)_time_subtracted

        desc: The timer named (name) just had some ticks removed.

        args:
            ticks: The new current tick number this timer is at.
            ticks_remaining: The new number of ticks in this timer remaining.
            ticks_subtracted: How many ticks were just subtracted from this
                timer. (This number will be positive, indicating the ticks
                subtracted.)
        '''

        self._check_for_done()

    def _check_for_done(self):
        # Checks to see if this timer is done. Automatically called anytime the
        # timer's value changes.

        self.debug_log("Checking to see if timer is done. Ticks: %s, End "
                       "Value: %s, Direction: %s",
                       self.ticks, self.end_value,
                       self.direction)

        if (self.direction == 'up' and self.end_value is not None and
                self.ticks >= self.end_value):
            self.timer_complete()
            return True

        if (self.direction == 'down' and
                self.ticks <= self.end_value):
            self.timer_complete()
            return True

        if self.end_value is not None:
            self.ticks_remaining = abs(self.end_value -
                                       self.ticks)

        self.debug_log("Timer is not done")

        return False

    def _create_system_timer(self):
        # Creates the clock event which drives this mode timer's tick method.
        self._remove_system_timer()
        self.timer = self.machine.clock.schedule_interval(self._timer_tick,
                                                          self.tick_secs)

    def _remove_system_timer(self):
        # Removes the clock event associated with this mode timer.
        if self.timer:
            self.machine.clock.unschedule(self.timer)
            self.timer = None

    @staticmethod
    def _get_timer_value(timer_value):
        if hasattr(timer_value, "evaluate"):
            # Convert to int for ticks; config_spec must be float for change_tick_interval
            return int(timer_value.evaluate([]))
        return timer_value

    def change_tick_interval(self, change=0.0, **kwargs):
        """Change the interval for each "tick" of this timer.

        Args:
        ----
            change: Float or int of the change you want to make to this timer's
                tick rate. Note this value is multiplied by the current tick
                interval: >1 will increase the tick interval (slow the timer) and
                <1 will decrease the tick interval (accelerate the timer).
                To set an absolute value, use the set_tick_interval() method.
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.tick_secs *= change.evaluate([])
        self._create_system_timer()

    def set_tick_interval(self, timer_value, **kwargs):
        """Set the number of seconds between ticks for this timer.

        This is an absolute setting. To apply a change to the current value,
        use the change_tick_interval() method.

        Args:
        ----
            timer_value: The new number of seconds between each tick of this
                timer. This value should always be positive.
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.tick_secs = abs(self._get_timer_value(timer_value))
        self._create_system_timer()

    def jump(self, timer_value, **kwargs):
        """Set the current amount of time of this timer.

        This value is expressed in "ticks" since the interval per tick can be
        something other than 1 second).

        Args:
        ----
            timer_value: Integer of the current value you want this timer to be.
            **kwargs: Not used in this method. Only exists since this method is
                often registered as an event handler which may contain
                additional keyword arguments.
        """
        del kwargs

        self.ticks = self._get_timer_value(timer_value)

        if self.max_value and self.ticks > self.max_value:
            self.ticks = self.max_value

        self._remove_system_timer()
        self._create_system_timer()

        self._check_for_done()

    def device_removed_from_mode(self, mode: Mode):
        """Stop this timer and also removes all the control events."""
        self.stop()
        self._remove_control_events()



"""Bit bang SPI to read switches."""
from typing import List

import asyncio
import logging

from mpf.core.utility_functions import Util

from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.core.platform import SwitchPlatform, SwitchConfig


class SpiBitBangSwitch(SwitchPlatformInterface):

    """Switch on SPI Bit Bang."""

    __slots__ = []  # type: List[str]

    def get_board_name(self):
        """Return board name."""
        return "SPI Big Bang"


class SpiBitBangPlatform(SwitchPlatform):

    """Platform which reads switch via SPI using bit banging."""

    __slots__ = ["_read_task", "config", "_switch_states"]

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Configure switch."""
        self._switch_states[number] = False
        return SpiBitBangSwitch(config, number)

    async def get_hw_switch_states(self):
        """Read initial hardware state.

        This will always be false for all inputs on those switches.
        """
        return self._switch_states

    def __init__(self, machine):
        """Initialise platform."""
        super().__init__(machine)
        self.log = logging.getLogger('SPI Bit Bang')
        self.log.debug("Configuring SPI Bit Bang.")
        self._read_task = None
        self.config = {}
        self._switch_states = {}

    def _enable_chip_select(self):
        self.config['cs_pin'].enable()

    def _disable_chip_select(self):
        self.config['cs_pin'].disable()

    async def initialize(self):
        """Register handler for late init."""
        self.machine.events.add_handler("init_phase_3", self._late_init)

    def _late_init(self, **kwargs):
        """Initialise this when other platforms are already loaded."""
        del kwargs
        self.config = self.machine.config_validator.validate_config("spi_bit_bang",
                                                                    self.machine.config.get('spi_bit_bang', {}))

        self._read_task = self.machine.clock.loop.create_task(self._run())
        self._read_task.add_done_callback(Util.raise_exceptions)

    async def read_spi(self, bits):
        """Read from SPI."""
        self.config['clock_pin'].disable()
        self._disable_chip_select()
        await asyncio.sleep(self.config['bit_time'])
        self._enable_chip_select()
        await asyncio.sleep(self.config['bit_time'])

        read_bits = 0
        for _ in range(bits):
            # read in bits on clk high
            read_bits <<= 1
            if self.config['miso_pin'].state:
                read_bits |= 0x1

            self.config['clock_pin'].pulse(int(self.config['clock_time']))
            await asyncio.sleep(self.config['bit_time'])

        self._disable_chip_select()
        return read_bits

    async def _run(self):
        while True:
            inputs = await self.read_spi(self.config['inputs'])
            for i in range(self.config['inputs']):
                num = str(i)
                if num in self._switch_states and self._switch_states[num] != bool(inputs & (1 << i)):
                    self._switch_states[num] = bool(inputs & (1 << i))
                    self.machine.switch_controller.process_switch_by_num(str(i), bool(inputs & (1 << i)), self)



"""Scriptlet.

DEPRECATED! Use custom code instead.
"""
from mpf.core.scriptlet import Scriptlet


class TestScriptlet(Scriptlet):
    def on_load(self):
        self.log.debug("Loaded!")

        self.machine.events.add_handler('test_event', self._update)

    def _update(self, **kwargs):
        del kwargs
        self.machine.events.post("test_response")
